/*
This source file is part of ProceduralGenerator (https://sourceforge.net/projects/proceduralgene/)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___DiffusionLimitedAggregation___
#define ___DiffusionLimitedAggregation___

#include <LibGenerator/ProceduralGenerator.h>
#include <set>

namespace ProceduralTextures
{
	struct DlaCell
	{
		UbPixel * m_pixel;
		bool m_whiteNeighbour;
		int m_iValue;
		bool m_white;

		DlaCell();
		DlaCell( const DlaCell & cell);
		~DlaCell();

		DlaCell & operator =( const DlaCell & cell);

		void Set( UbPixel * p_pixel, bool p_white);
		void SetWhite( double p_dDistance);
		bool IncValue( int p_iMinValue);
	};
}

namespace ProceduralTextures
{
	template <> class Policy<ProceduralTextures::DlaCell>
	{
	private:
		typedef ProceduralTextures::DlaCell value_type;

	public:
		static value_type	zero			()							{ return value_type(); }
		static value_type	unit			()							{ return value_type(); }
		static void			init			( value_type & p_a)			{ p_a = zero(); }
		static bool			is_null			( const value_type & p_a)	{ return equals( p_a, zero()); }
		static value_type	negate			( const value_type & p_a)	{ return p_a; }
		static value_type	ass_negate		( value_type & p_a)			{ return assign( p_a, negate( p_a)); }
		static value_type	stick			( value_type & p_a)			{}

		template <typename Ty> static value_type	convert			( const Ty & p_value)						{ return static_cast<value_type>( p_value); }
		template <typename Ty> static bool			equals			( const value_type & p_a, const Ty & p_b)	{ return false; }
		template <typename Ty> static value_type	add				( const value_type & p_a, const Ty & p_b)	{ return p_a; }
		template <typename Ty> static value_type	substract		( const value_type & p_a, const Ty & p_b)	{ return p_a; }
		template <typename Ty> static value_type	multiply		( const value_type & p_a, const Ty & p_b)	{ return p_a; }
		template <typename Ty> static value_type	divide			( const value_type & p_a, const Ty & p_b)	{ return p_a; }
		template <typename Ty> static value_type	ass_add			( value_type & p_a, const Ty & p_b)			{ return p_a; }
		template <typename Ty> static value_type	ass_substract	( value_type & p_a, const Ty & p_b)			{ return p_a; }
		template <typename Ty> static value_type	ass_multiply	( value_type & p_a, const Ty & p_b)			{ return p_a; }
		template <typename Ty> static value_type	ass_divide		( value_type & p_a, const Ty & p_b)			{ return p_a; }
		template <typename Ty> static value_type	assign			( value_type & p_a, const Ty & p_b)			{ return p_a = convert<Ty>( p_b); }
	};
}

namespace ProceduralTextures
{
	typedef Buffer<DlaCell> DlaCellBuffer;

	class DiffusionLimitedAggregation : public ProceduralGenerator
	{
	private:
		class Thread : public ProceduralGenerator::Thread
		{
		public:
			DlaCellBuffer * m_cells;
			std::vector <bool> m_passed;
			int m_iKillCircle;
			int m_iJumpCircle;
			int m_iBoundingCircle;
			Point<int, 2> m_ptOrigin;
			int m_iMinValue;
			int * m_iCurrentMass;
			double m_dMaxDistance;

		public:
			Thread( ProceduralGenerator * p_pParent, size_t p_uiIndex, int iWidth, int iTop, int iBottom, int iTotalHeight, const UbPixel & p_pxColour);
			virtual ~Thread();

			virtual void Step();

			inline void SetBuffer( DlaCellBuffer * p_cells) { m_cells = p_cells; }

			void _getPointOnCircle( Point<int, 2> & p_ptPoint, const Point<int, 2> & p_ptOrigin, float p_fRadius);
			void _setCellWhite( const Point<int, 2> & p_ptPoint);
		};

	public:
		DlaCellBuffer m_cells;
		float m_fStickCoeff;
		int m_iKillCircle;
		int m_iJumpCircle;
		int m_iBoundingCircle;
		Point<int, 2> m_ptOrigin;
		int m_iMinValue;
		int m_iMass;
		int m_iCurrentMass;
		double m_dMaxDistance;

	public:
		DiffusionLimitedAggregation( int p_width, int p_height, int p_iWndId, wxFrame * p_pFrame);
		virtual ~DiffusionLimitedAggregation();

		virtual void SwapBuffers();

		inline int	GetMinValue	()const { return m_iMinValue; }
		inline int	GetMass		()const { return m_iMass; }

		inline void SetMass	( int val)		{ m_iMass = val; }
		inline void SetMinValue	( int val)	{ m_iMinValue = val; }

	private:
		virtual void _glInit		(){}
		virtual void _glPreRender	(){}
		virtual void _glRender		( bool & p_bChanged){}
		virtual void _glPostRender	(){}
		virtual void _glCleanup		(){}

		void _setCellWhite( const Point<int, 2> & p_ptPoint);
	};

	class DlaPanel : public ConfigPanel
	{
	private:
		typedef enum
		{	eMass		= 50
		,	eMinValue	= 51
		}
		eIDs;

		DiffusionLimitedAggregation	*	m_pGenerator;
		wxSlider					*	m_pSliderMass;
		wxSlider					*	m_pSliderMinValue;
		wxStaticText				*	m_pStaticMass;
		wxStaticText				*	m_pStaticMinValue;

	public:
		DlaPanel( wxWindow * p_pWindow, wxWindowID p_id, const wxPoint & p_ptPosition=wxDefaultPosition, const wxSize & p_size=wxDefaultSize);
		virtual ~DlaPanel();

		void Create( DiffusionLimitedAggregation * p_pGenerator, wxWindow * p_pWindow, const wxPoint & p_ptPosition, const wxSize & p_size);

	private:
		DECLARE_EVENT_TABLE()
		void _onSize		( wxSizeEvent		&	p_event);
		void _onMinValue	( wxCommandEvent	&	p_event);
		void _onMass		( wxCommandEvent	&	p_event);
	};

	class DlaPlugin : public Plugin<DlaPanel, DiffusionLimitedAggregation, DlaPlugin>
	{
	public:
		DlaPlugin();
		virtual ~DlaPlugin();

		virtual wxString GetName();

		DECLARE_PLUGIN()
	};
}

#endif
