/*
This source file is part of ProceduralGenerator (https://sourceforge.net/projects/proceduralgene/)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___PG_Point___
#define ___PG_Point___

#include "Value.h"
#include <iostream>

namespace ProceduralTextures
{
	/*!
	Base class for points, used to share functions without taking care of templates
	\author Sylvain DOREMUS
	\version 0.6.1.0
	\date 03/01/2011
	*/
	class PG_API PointBase
	{
	public:
		/**@name Construction / Destruction */
		//@{
		PointBase(){}
		virtual ~PointBase(){}
		//@}

		/**@name Copy */
		//@{
		PointBase( const PointBase & p_ptBase){}
		virtual PointBase & operator =( const PointBase & p_ptBase){ return * this; }
		//@}

		virtual size_t			GetElementCount()const=0;
		virtual size_t			GetElementSize()const=0;
		virtual void			LinkCoords( const void * p_pCoords)=0;
		virtual void 			UnlinkCoords()=0;

		template <typename T> T			*	ptr			()		{ return (T *)_ptr(); }
		template <typename T> const T	*	const_ptr	()const	{ return (const T *)_const_ptr(); }

	protected:
		virtual void		*	_ptr() { return NULL; }
		virtual const void	*	_const_ptr()const { return NULL; }
	};

	/*!
	Templated dynamic dimensions point class
	\author Sylvain DOREMUS
	\version 0.6.1.0
	\date 03/01/2011
	*/
	template <typename T>
	class DynPoint : public PointBase
	{
	private:
		template <typename U> friend class DynPoint;

		typedef T								value_type;
		typedef value_type &					reference;
		typedef value_type *					pointer;
		typedef const value_type &				const_reference;
		typedef const value_type *				const_pointer;
		typedef DynPoint<value_type>			point;
		typedef DynPoint<value_type> &			point_reference;
		typedef DynPoint<value_type> *			point_pointer;
		typedef const DynPoint<value_type> &	const_point_reference;
		typedef const DynPoint<value_type> *	const_point_pointer;

		size_t m_uiCount;
		bool m_bOwnCoords;
		T * m_coords;

	public:
		/**@name Construction / Destruction */
		//@{
		DynPoint( size_t p_uiCount);
		template <typename U> DynPoint( size_t p_uiCount, const U * p_pValues);
		template <typename U> DynPoint( const U & p_vA, const U & p_vB);
		template <typename U> DynPoint( const U & p_vA, const U & p_vB, const U & p_vC);
		template <typename U> DynPoint( const U & p_vA, const U & p_vB, const U & p_vC, const U & p_vD);
		template <typename U> DynPoint( const DynPoint<U> & p_ptPoint);
		virtual ~DynPoint();
		//@}

		//*@name Copy */
		//@{
		DynPoint( const DynPoint<T> & p_ptPoint);
		DynPoint<T> & operator =( const DynPoint<T> & p_pt);
		//@}

		/**@name Operators */
		//@{
		template <typename U> DynPoint<T> & operator =	( const DynPoint<U> & p_pt);
		template <typename U> DynPoint<T> & operator +=	( const DynPoint<U>	& p_pt);
		template <typename U> DynPoint<T> & operator -=	( const DynPoint<U>	& p_pt);
		template <typename U> DynPoint<T> & operator *=	( const DynPoint<U>	& p_pt);
		template <typename U> DynPoint<T> & operator /=	( const DynPoint<U>	& p_pt);
		template <typename U> DynPoint<T> & operator ^=	( const DynPoint<U>	& p_pt);
		template <typename U> DynPoint<T> & operator +=	( const U			* p_coords);
		template <typename U> DynPoint<T> & operator -=	( const U			* p_coords);
		template <typename U> DynPoint<T> & operator *=	( const U			* p_coords);
		template <typename U> DynPoint<T> & operator /=	( const U			* p_coords);
		template <typename U> DynPoint<T> & operator +=	( const U			& p_coord);
		template <typename U> DynPoint<T> & operator -=	( const U			& p_coord);
		template <typename U> DynPoint<T> & operator *=	( const U			& p_coord);
		template <typename U> DynPoint<T> & operator /=	( const U			& p_coord);
		//@}

		template <typename U>
		void					swap				( const DynPoint<U> & p_pt);

		/**@name Accessors */
		//@{
		virtual size_t			GetElementCount		()const						{ return m_uiCount; }
		virtual size_t			GetElementSize		()const						{ return sizeof( T); }
		virtual void			LinkCoords			( const void * p_pCoords);
		virtual void			UnlinkCoords		();
		void 					ToValues			( T * p_pResult)const;
		inline const T		&	operator[]			( size_t p_pos)const		{ return m_coords[p_pos]; }
		inline T			&	operator[]			( size_t p_pos)				{ return m_coords[p_pos]; }
		inline const T		&	at					( size_t p_pos)const		{ return m_coords[p_pos]; }
		inline T			&	at					( size_t p_pos)				{ return m_coords[p_pos]; }
		inline T			*	ptr					()							{ return m_coords; }
		inline const T		*	const_ptr			()const						{ return m_coords; }

	private:
		inline void			*	_ptr				()							{ return m_coords; }
		inline const void	*	_const_ptr			()const						{ return m_coords; }
		//@}
	};

	template <typename T, typename U	> bool				operator ==	( const DynPoint<T>		& p_ptA,		const	DynPoint<U>	& p_ptB);
	template <typename T, typename U	> bool				operator !=	( const DynPoint<T>		& p_ptA,		const	DynPoint<U>	& p_ptB);
	template <typename T				> std::ostream	&	operator <<	(		std::ostream	& p_streamOut,	const	DynPoint<T>	& p_pt);
	template <typename T				> std::istream	&	operator >>	(		std::istream	& p_streamIn,			DynPoint<T>	& p_pt);
	template <typename T, typename U	> DynPoint<T>		operator +	( const DynPoint<T>		& p_ptA,		const	DynPoint<U>	& p_ptB);
	template <typename T, typename U	> DynPoint<T>		operator -	( const DynPoint<T>		& p_ptA,		const	DynPoint<U>	& p_ptB);
	template <typename T, typename U	> DynPoint<T>		operator *	( const DynPoint<T>		& p_ptA,		const	DynPoint<U>	& p_ptB);
	template <typename T, typename U	> DynPoint<T>		operator /	( const DynPoint<T>		& p_ptA,		const	DynPoint<U>	& p_ptB);
	template <typename T, typename U	> DynPoint<T>		operator ^	( const DynPoint<T>		& p_ptA,		const	DynPoint<U>	& p_ptB);
	template <typename T, typename U	> DynPoint<T>		operator +	( const DynPoint<T>		& p_pt,			const	U			* p_coords);
	template <typename T, typename U	> DynPoint<T>		operator -	( const DynPoint<T>		& p_pt,			const	U			* p_coords);
	template <typename T, typename U	> DynPoint<T>		operator *	( const DynPoint<T>		& p_pt,			const	U			* p_coords);
	template <typename T, typename U	> DynPoint<T>		operator /	( const DynPoint<T>		& p_pt,			const	U			* p_coords);
	template <typename T, typename U	> DynPoint<T>		operator +	( const DynPoint<T>		& p_pt,			const	U			& p_coord);
	template <typename T, typename U	> DynPoint<T>		operator -	( const DynPoint<T>		& p_pt,			const	U			& p_coord);
	template <typename T, typename U	> DynPoint<T>		operator *	( const DynPoint<T>		& p_pt,			const	U			& p_coord);
	template <typename T, typename U	> DynPoint<T>		operator /	( const DynPoint<T>		& p_pt,			const	U			& p_coord);

	template <typename T				> DynPoint<T>		operator *	( int p_value,							const	DynPoint<T>	& p_pt);
	template <typename T				> DynPoint<T>		operator +	( int p_value,							const	DynPoint<T>	& p_pt);
	template <typename T				> DynPoint<T>		operator -	( int p_value,							const	DynPoint<T>	& p_pt);
	template <typename T				> DynPoint<T>		operator /	( int p_value,							const	DynPoint<T>	& p_pt);
	template <typename T				> DynPoint<T>		operator *	( double p_value,						const	DynPoint<T>	& p_pt);
	template <typename T				> DynPoint<T>		operator +	( double p_value,						const	DynPoint<T>	& p_pt);
	template <typename T				> DynPoint<T>		operator -	( double p_value,						const	DynPoint<T>	& p_pt);
	template <typename T				> DynPoint<T>		operator /	( double p_value,						const	DynPoint<T>	& p_pt);

	//! Templated static dimensions point representation
	/*!
	Can hold any type which has a defined Policy
	\author Sylvain DOREMUS
	\date 14/02/2010
	*/
	template <typename T, size_t Count>
	class Point : public PointBase
	{
	private:
		template <typename U, size_t _Count> friend class Point;

		typedef T									value_type;
		typedef value_type &						reference;
		typedef value_type *						pointer;
		typedef const value_type &					const_reference;
		typedef const value_type *					const_pointer;
		typedef Point<value_type, Count>			point;
		typedef Point<value_type, Count> &			point_reference;
		typedef Point<value_type, Count> *			point_pointer;
		typedef const Point<value_type, Count> &	const_point_reference;
		typedef const Point<value_type, Count> *	const_point_pointer;

		bool m_bOwnCoords;
		T * m_coords;

	public:
		/**@name Constructors */
		//@{
		Point();
		template <typename U> Point( const U * p_pValues);
		template <typename U> Point( const U & p_vA);
		template <typename U> Point( const U & p_vA, const U & p_vB);
		template <typename U> Point( const U & p_vA, const U & p_vB, const U & p_vC);
		template <typename U> Point( const U & p_vA, const U & p_vB, const U & p_vC, const U & p_vD);
		template <typename U, size_t _Count> Point( const Point<U, _Count> & p_ptPoint);
		virtual ~Point();
		//@}

		//*@name Copy */
		//@{
		Point( const Point<T, Count> & p_ptPoint);
		Point<T, Count> & operator =( const Point<T, Count> & p_pt);
		//@}

		/**@name Operators */
		//@{
		template <typename U, size_t _Count	> Point <T, Count> & operator =		( const Point<U, _Count>	& p_pt);
		template <typename U, size_t _Count	> Point <T, Count> & operator +=	( const Point<U, _Count>	& p_pt);
		template <typename U, size_t _Count	> Point <T, Count> & operator -=	( const Point<U, _Count>	& p_pt);
		template <typename U, size_t _Count	> Point <T, Count> & operator *=	( const Point<U, _Count>	& p_pt);
		template <typename U, size_t _Count	> Point <T, Count> & operator /=	( const Point<U, _Count>	& p_pt);
		template <typename U, size_t _Count	> Point <T, Count> & operator ^=	( const Point<U, _Count>	& p_pt);
		template <typename U				> Point <T, Count> & operator +=	( const U					* p_coords);
		template <typename U				> Point <T, Count> & operator -=	( const U					* p_coords);
		template <typename U				> Point <T, Count> & operator *=	( const U					* p_coords);
		template <typename U				> Point <T, Count> & operator /=	( const U					* p_coords);
		template <typename U				> Point <T, Count> & operator +=	( const U					& p_coord);
		template <typename U				> Point <T, Count> & operator -=	( const U					& p_coord);
		template <typename U				> Point <T, Count> & operator *=	( const U					& p_coord);
		template <typename U				> Point <T, Count> & operator /=	( const U					& p_coord);
		//@}

		/**@name Operations */
		//@{
		void 					Reverse				();
		void 					Normalise			();
		point 					GetNormalised		()const;
		value_type 				Dot					( const_point_reference p_vertex)const;
		double 					GetSquaredLength	()const;
		double 					GetLength			()const;
		double 					GetManhattanLength	()const;
		double 					GetMinkowskiLength	( double p_dOrder)const;
		double 					GetChebychevLength	()const;
		double 					GetCosTheta			( const_point_reference p_vector)const;
		void					LinkCoords			( const void * p_pCoords);
		void					UnlinkCoords		();
		template <typename U, size_t _Count	>
		void					swap				( const Point<U, _Count> & p_pt);
		//@}

		/**@name Accessors */
		//@{
		void 					ToValues		( T * p_pResult)const;
		inline size_t			GetElementCount	()const					{ return Count; }
		inline size_t			GetElementSize	()const					{ return sizeof( T); }
		inline const T		&	operator[]		( size_t p_pos)const	{ return m_coords[p_pos]; }
		inline T			&	operator[]		( size_t p_pos)			{ return m_coords[p_pos]; }
		inline const T		&	at				( size_t p_pos)const	{ return m_coords[p_pos]; }
		inline T			&	at				( size_t p_pos)			{ return m_coords[p_pos]; }
		inline T			*	ptr				()						{ return m_coords; }
		inline const T		*	const_ptr		()const					{ return m_coords; }

	private:
		inline void			*	_ptr			()						{ return m_coords; }
		inline const void	*	_const_ptr		()const					{ return m_coords; }
		//@}
	};

	template <typename T, size_t Count, typename U, size_t _Count	> bool					operator ==	( const Point<T, Count> & p_ptA,		const	Point<U, _Count>	& p_ptB);
	template <typename T, size_t Count, typename U, size_t _Count	> bool					operator !=	( const Point<T, Count> & p_ptA,		const	Point<U, _Count>	& p_ptB);
	template <typename T, size_t Count								> std::ostream		&	operator <<	(		std::ostream	& p_streamOut,	const	Point<T, Count>		& p_pt);
	template <typename T, size_t Count								> std::istream		&	operator >>	(		std::istream	& p_streamIn,			Point<T, Count>		& p_pt);
	template <typename T, size_t Count, typename U, size_t _Count	> Point <T, Count>		operator +	( const Point<T, Count> & p_ptA,		const	Point<U, _Count>	& p_ptB);
	template <typename T, size_t Count, typename U, size_t _Count	> Point <T, Count>		operator -	( const Point<T, Count> & p_ptA,		const	Point<U, _Count>	& p_ptB);
	template <typename T, size_t Count, typename U, size_t _Count	> Point <T, Count>		operator *	( const Point<T, Count> & p_ptA,		const	Point<U, _Count>	& p_ptB);
	template <typename T, size_t Count, typename U, size_t _Count	> Point <T, Count>		operator /	( const Point<T, Count> & p_ptA,		const	Point<U, _Count>	& p_ptB);
	template <typename T, size_t Count, typename U, size_t _Count	> Point <T, Count>		operator ^	( const Point<T, Count> & p_ptA,		const	Point<U, _Count>	& p_ptB);
	template <typename T, size_t Count, typename U					> Point <T, Count>		operator +	( const Point<T, Count> & p_pt,			const	U					* p_coords);
	template <typename T, size_t Count, typename U					> Point <T, Count>		operator -	( const Point<T, Count> & p_pt,			const	U					* p_coords);
	template <typename T, size_t Count, typename U					> Point <T, Count>		operator *	( const Point<T, Count> & p_pt,			const	U					* p_coords);
	template <typename T, size_t Count, typename U					> Point <T, Count>		operator /	( const Point<T, Count> & p_pt,			const	U					* p_coords);
	template <typename T, size_t Count, typename U					> Point <T, Count>		operator +	( const Point<T, Count> & p_pt,			const	U					& p_coord);
	template <typename T, size_t Count, typename U					> Point <T, Count>		operator -	( const Point<T, Count> & p_pt,			const	U					& p_coord);
	template <typename T, size_t Count, typename U					> Point <T, Count>		operator *	( const Point<T, Count> & p_pt,			const	U					& p_coord);
	template <typename T, size_t Count, typename U					> Point <T, Count>		operator /	( const Point<T, Count> & p_pt,			const	U					& p_coord);

	template <typename T, size_t Count								> Point <T, Count>		operator *	( int p_value,							const	Point <T, Count>	& p_pt);
	template <typename T, size_t Count								> Point <T, Count>		operator +	( int p_value,							const	Point <T, Count>	& p_pt);
	template <typename T, size_t Count								> Point <T, Count>		operator -	( int p_value,							const	Point <T, Count>	& p_pt);
	template <typename T, size_t Count								> Point <T, Count>		operator /	( int p_value,							const	Point <T, Count>	& p_pt);
	template <typename T, size_t Count								> Point <T, Count>		operator *	( double p_value,						const	Point <T, Count>	& p_pt);
	template <typename T, size_t Count								> Point <T, Count>		operator +	( double p_value,						const	Point <T, Count>	& p_pt);
	template <typename T, size_t Count								> Point <T, Count>		operator -	( double p_value,						const	Point <T, Count>	& p_pt);
	template <typename T, size_t Count								> Point <T, Count>		operator /	( double p_value,						const	Point <T, Count>	& p_pt);

	typedef Point<float,	1>	Point1f;
	typedef Point<float,	2>	Point2f;
	typedef Point<float,	3>	Point3f;
	typedef Point<float,	4>	Point4f;
	typedef Point<int,		1>	Point1i;
	typedef Point<int,		2>	Point2i;
	typedef Point<int,		3>	Point3i;
	typedef Point<int,		4>	Point4i;
	typedef Point<byte,		1>	Point1ub;
	typedef Point<byte,		2>	Point2ub;
	typedef Point<byte,		3>	Point3ub;
	typedef Point<byte,		4>	Point4ub;
}

#include "Point.inl"

#endif
