/*
This source file is part of ProceduralGenerator (https://sourceforge.net/projects/proceduralgene/)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___ShaderEffects___
#define ___ShaderEffects___

#include <LibGenerator/glew.h>
#include <wx/image.h>
#include <wx/glcanvas.h>
#include <wx/notebook.h>
#include <LibGenerator/ProceduralGenerator.h>
#include <LibGenerator/GlObjects.h>

namespace ProceduralTextures
{
	class SeEffectsNotebook;
	class SeEffectPage;
	class ShaderEffects;
	class SePlugin;
	class SePanel;

	class ShaderEffects : public ProceduralGenerator
	{
	public:
		typedef enum
		{	eSEPARATION_NONE
		,	eSEPARATION_HORIZ
		,	eSEPARATION_VERT
		}	eSEPARATION;

	private:
		class Effect
		{
		private:
			wxImage *					m_pImage;
			std::vector <wxImage *>		m_arrayImages;
			std::vector <wxString>		m_arrayImagesFiles;
			GlShaderProgram			*	m_pShaderProgram;
			wxString					m_strVertexFile;
			wxString					m_strFragmentFile;
			bool						m_bNewShader;
			wxMilliClock_t				m_llStartTime;
			int							m_iWidth;
			int							m_iHeight;

		public:
			Effect( wxImage * p_pImage);
			virtual ~Effect();

			void SetVertexFile( const wxString & p_strPath);
			void SetFragmentFile( const wxString & p_strPath);
			void Compile();
			void SetImagePath( size_t p_iImage, const wxString & p_strImagePath);

			void GlPreRender();
			void GlPostRender();
			void GlCleanup();

			bool Activate( eSEPARATION p_eSeparationType, int p_iSeparationOffset);
			void Deactivate();

			wxString GetCompilerLog( eSHADER_OBJECT_TYPE p_eType);
			wxString GetLinkerLog();

			void UpdateImages();

			inline void ResetTime	()	{ m_llStartTime = wxGetLocalTimeMillis(); }
		};

	private:
		wxImage					m_image;
		std::vector<Effect *>	m_arrayEffects;
		eSEPARATION				m_eSeparationType;
		int						m_iSeparationOffset;
		GlFrameBuffer			m_frameBuffer1;
		GlFrameBuffer			m_frameBuffer2;
		GlTexture				m_texture1;
		GlTexture				m_texture2;

	public:
		ShaderEffects( int p_width, int p_height, int p_iWndId, wxFrame * p_pFrame);
		virtual ~ShaderEffects();

		void SetImagePath	( const wxString & p_strImagePath);
		void ResetTime		();

		bool		EffectAdd				();
		bool		EffectRemove			( size_t p_uiIndex);
		void		EffectSetVertexFile		( size_t p_uiIndex, const wxString & p_strPath);
		void		EffectSetFragmentFile	( size_t p_uiIndex, const wxString & p_strPath);
		void		EffectSetImagePath		( size_t p_uiIndex, size_t p_iImage, const wxString & p_strImagePath);
		void		EffectCompile			( size_t p_uiIndex);
		wxString	EffectGetCompilerLog	( size_t p_uiIndex, eSHADER_OBJECT_TYPE p_eType);
		wxString	EffectGetLinkerLog		( size_t p_uiIndex);
		bool		EffectActivate			( size_t p_uiIndex);
		void		EffectDeactivate		( size_t p_uiIndex);

		inline void SetSeparationOffset	( int p_iOffset)		{ m_iSeparationOffset = p_iOffset; }
		inline void SetSeparationType	( eSEPARATION p_eType)	{ m_eSeparationType = p_eType; }

	private:
		virtual void _resize		( const wxSize & p_size);
		virtual void _glInit		();
		virtual void _glPreRender	();
		virtual void _glRender		( bool & p_bChanged);
		virtual void _glPostRender	();
		virtual void _glCleanup		();
	};

	class SePanel : public ConfigPanel
	{
	public:
		typedef enum
		{	eImagePath		= 52
		,	eSepType		= 53
		,	eSepOffset		= 54
		,	eResetTime		= 55
		,	eShaders		= 56
		} eIDs;

	private:
		ShaderEffects		*	m_pGenerator;
		wxButton			*	m_pButtonImage;
		wxComboBox			*	m_pComboSeparationType;
		wxSlider			*	m_pSliderSeparationOffset;
		wxButton			*	m_pButtonResetTime;
		SeEffectsNotebook	*	m_pShadersConfigs;

	public:
		SePanel( wxWindow * p_pWindow, wxWindowID p_id, const wxPoint & p_ptPosition=wxDefaultPosition, const wxSize & p_size=wxDefaultSize);
		virtual ~SePanel();

		void Create( ShaderEffects * p_pGenerator, wxWindow * p_pWindow, const wxPoint & p_ptPosition, const wxSize & p_size);

	private:
		DECLARE_EVENT_TABLE();
		void _onSize				( wxSizeEvent		&	p_event);
		void _onImagePath			( wxCommandEvent	&	p_event);
		void _onSepType				( wxCommandEvent	&	p_event);
		void _onSepOffset			( wxCommandEvent	&	p_event);
		void _onResetTime			( wxCommandEvent	&	p_event);
	};

	class SeEffectPage : public wxNotebookPage
	{
	private:
		typedef enum
		{	eVertexFile		= 60
		,	eFragmentFile	= 61
		,	eCompile		= 62
		,	eCompilerLog	= 63
		,	eRemove			= 64
		} eIDs;

		SeEffectsNotebook	*	m_pParent;
		ShaderEffects		*	m_pGenerator;
		wxToggleButton		*	m_pButtonVertexFile;
		wxToggleButton		*	m_pButtonFragmentFile;
		wxButton			*	m_pButtonCompile;
		wxButton			*	m_pButtonCompilerLog;
		wxButton			*	m_pButtonRemove;
		size_t					m_uiIndex;

	public:
		SeEffectPage( SeEffectsNotebook * p_pParent, ShaderEffects * p_pGenerator, size_t p_uiIndex, bool p_bCanRemove, const wxPoint & p_ptPosition, const wxSize & p_size);
		virtual ~SeEffectPage();

		inline void DecIndex() { m_uiIndex--; }

	private:
		DECLARE_EVENT_TABLE();
		void _onVertexShaderPath	( wxCommandEvent	&	p_event);
		void _onFragmentShaderPath	( wxCommandEvent	&	p_event);
		void _onShaderCompile		( wxCommandEvent	&	p_event);
		void _onCompilerLog			( wxCommandEvent	&	p_event);
		void _onRemove				( wxCommandEvent	&	p_event);
	};

	class SeEffectsNotebook : public wxNotebook
	{
	private:
		std::vector <SeEffectPage *>	m_listPages;
		SeEffectPage				*	m_pSelection;
		ShaderEffects				*	m_pGenerator;

	public:
		SeEffectsNotebook( wxWindow * p_pParent, wxWindowID p_id, const wxPoint & p_ptPosition=wxDefaultPosition, const wxSize & p_size=wxDefaultSize, long lStyle=0L);
		virtual ~SeEffectsNotebook();

		SeEffectPage * CreatePage( ShaderEffects * p_pGenerator, const wxString & p_strTitle);

		virtual bool DeletePage( size_t p_uiPage);

	private:
		DECLARE_EVENT_TABLE();
		void _onSize( wxSizeEvent & p_event);
		void _onSelectPage( wxBookCtrlEvent & p_event);
	};

	class SePlugin : public Plugin<SePanel, ShaderEffects, SePlugin>
	{
	public:
		SePlugin();
		virtual ~SePlugin();

		virtual wxString GetName();

		DECLARE_PLUGIN()
	};
}

#endif
