/*
This source file is part of ProceduralGenerator (https://sourceforge.net/projects/proceduralgene/)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___VoronoiTexture___
#define ___VoronoiTexture___

#include <LibGenerator/ProceduralGenerator.h>
#include <vector>
#include <set>
#include <map>

namespace ProceduralTextures
{
	class Voronoi2D : public ProceduralGenerator
	{
	private:
		class Seed;
		typedef std::vector <Seed>			SeedArray;
		typedef std::vector <Seed *>		SeedPtrArray;
		typedef std::vector <SeedPtrArray>	SeedPtrMatrix;


		class Seed : public Point<int, 2>
		{
		private:
			UbPixel				m_pxColour;
			PixelBuffer		*	m_pBuffer;
			bool				m_bCanGrow;
			SeedPtrMatrix	*	m_pMatrix;
			Point<int, 2>		m_ptMin;
			Point<int, 2>		m_ptMax;

		public:
			Seed( int p_iX, int p_iY, UbPixel const & p_pxColour, PixelBuffer * p_pBuffer, SeedPtrMatrix * p_pMatrix);
			Seed( Seed const & p_seed);
			virtual ~Seed();

			Seed & operator =( Seed const & p_seed);

			void Grow();
		};

	public:
		typedef Point		<unsigned char	, 3	>	Colour;
		typedef std::vector	<Colour				>	ColourArray;
		typedef std::set	<int				>	IntSet;
		typedef std::map	<int, int			>	IntIntMap;
		typedef std::map	<int, IntSet		>	IntSetIntMap;
		typedef std::map	<int, IntIntMap		>	IntIntMapIntMap;


		typedef enum
		{	eDISTANCE_FUNC_EUCLIDIAN
		,	eDISTANCE_FUNC_SQUARED_EUCLIDIAN
		,	eDISTANCE_FUNC_MANHATTAN
		,	eDISTANCE_FUNC_MINKOWSKI
		,	eDISTANCE_FUNC_CHEBYCHEV
		,	eDISTANCE_FUNC_COUNT
		}	eDISTANCE_FUNC;

	private:
		class Thread : public ProceduralGenerator::Thread
		{
		private:
			typedef int (Voronoi2D::Thread::*DistanceFunction)( const Point<int, 2> & p_ptA, const Point<int, 2> & p_ptB);
			PixelBuffer			*	m_pBackBuffer;
			IntSetIntMap		*	m_pSeeds;
			IntIntMapIntMap		*	m_pHeights;
			ColourArray			*	m_pArrayColours;
			ColourArray			*	m_pArrayStepColours;
			eDISTANCE_FUNC			m_eTypeDistance;
			int						m_iMaxDepth;
			int						m_iMaxNeighbours;
			int						m_iX;
			int						m_iY;
			double					m_dMkOrder;
			DistanceFunction		m_pFunctions[eDISTANCE_FUNC_COUNT];

		public:
			Thread( ProceduralGenerator * p_pParent, size_t p_uiIndex, int iWidth, int iTop, int iBottom, int iTotalHeight, const UbPixel & p_pxColour);
			virtual ~Thread();

			virtual void Step();

			void Initialise( PixelBuffer * p_pBackBuffer, IntSetIntMap * p_pSeeds, IntIntMapIntMap * p_pHeights, ColourArray * p_pArrayColours, ColourArray * p_pArrayStepColours, eDISTANCE_FUNC p_eTypeDistance, int p_iMaxDepth, int p_iMaxNeighbours, double p_dMkOrder);
			void ComputeColour( int x, int y, UbPixel & p_px);

			int EuclidianLength( const Point<int, 2> & p_ptA, const Point<int, 2> & p_ptB);
			int SquaredEuclidianLength( const Point<int, 2> & p_ptA, const Point<int, 2> & p_ptB);
			int ManhattanLength( const Point<int, 2> & p_ptA, const Point<int, 2> & p_ptB);
			int MinkowskiLength( const Point<int, 2> & p_ptA, const Point<int, 2> & p_ptB);
			int ChebychevLength( const Point<int, 2> & p_ptA, const Point<int, 2> & p_ptB);
		};

	private:
		IntSetIntMap	m_seeds;
		IntIntMapIntMap	m_heights;
		IntSet			m_setX;
		IntSet			m_setY;
		int				m_iMaxDepth;
		int				m_iMaxNeighbours;
		ColourArray		m_arrayColours;
		ColourArray		m_arrayStepColours;
		eDISTANCE_FUNC	m_eTypeDistance;
		double			m_dMkOrder;

	public:
		Voronoi2D( int p_width, int p_height, int p_iWndId, wxFrame * p_pFrame);
		virtual ~Voronoi2D();

		void SetMaxDepth( int p_iMaxDepth);
		void SetMaxNeighbours( int p_iMaxNeighbours);
		void SetTypeDistance( eDISTANCE_FUNC p_eType);
		void SetMinkowskiOrder( double p_dMkOrder);

		void Initialise();
		void Generate();

		inline int				GetMaxDepth			()const { return m_iMaxDepth; }
		inline int				GetMaxNeighbours	()const { return m_iMaxNeighbours; }
		inline eDISTANCE_FUNC	GetTypeDistance		()const	{ return m_eTypeDistance; }
		inline double			GetMinkowskiOrder	()const	{ return m_dMkOrder; }

	private:
		virtual void _glInit		(){}
		virtual void _glPreRender	(){}
		virtual void _glRender		( bool & p_bChanged){}
		virtual void _glPostRender	(){}
		virtual void _glCleanup		(){}
	};

	class V2dPanel : public ConfigPanel
	{
	private:
		typedef enum
		{	eID_DEPTH			= 50
		,	eID_NEIGHBOURS		= 51
		,	eID_DISTANCE_FUNC	= 52
		,	eID_MINKOWSKI_ORDER	= 53
		}	eID;

		Voronoi2D		*	m_pGenerator;
		wxStaticText	*	m_pStaticMaxDepths;
		wxSlider		*	m_pSliderMaxDepths;
		wxStaticText	*	m_pStaticMaxNeighbours;
		wxSlider		*	m_pSliderMaxNeighbours;
		wxComboBox		*	m_pComboDistanceFunction;
		wxStaticText	*	m_pStaticMinkowskiOrder;
		wxSlider		*	m_pSliderMinkowskiOrder;

	public:
		V2dPanel( wxWindow * p_pWindow, wxWindowID p_id, const wxPoint & p_ptPosition=wxDefaultPosition, const wxSize & p_size=wxDefaultSize);
		virtual ~V2dPanel();

		void Create( Voronoi2D * p_pGenerator, wxWindow * p_pWindow, const wxPoint & p_ptPosition, const wxSize & p_size);

	private:
		DECLARE_EVENT_TABLE()
		void _onMaxDepths			( wxScrollEvent		&	p_event);
		void _onMaxNeighbours		( wxScrollEvent		&	p_event);
		void _onMkOrder				( wxScrollEvent		&	p_event);
		void _onDistanceType		( wxCommandEvent	&	p_event);
		void _onMaxDepthsTrack		( wxScrollEvent		&	p_event);
		void _onMaxNeighboursTrack	( wxScrollEvent		&	p_event);
		void _onMkOrderTrack		( wxScrollEvent		&	p_event);
	};

	class V2dPlugin : public Plugin<V2dPanel, Voronoi2D, V2dPlugin>
	{
	public:
		V2dPlugin();
		virtual ~V2dPlugin();

		virtual wxString GetName();

		DECLARE_PLUGIN()
	};
}

#endif
