#include "ImageEffects/GeneratorImageEffects.h"

#include <wx/sysopt.h>
#include <wx/display.h>
#include <wx/tglbtn.h>
#include <cmath>

#ifdef __WXMSW__
#	include<wx/msw/msvcrt.h>
#endif

using namespace ProceduralTextures;

//*************************************************************************************************

ImageEffects::EffectFunctor :: EffectFunctor( eEFFECT_TYPE	p_eEffectType, const wxString & p_strName, bool p_bAllowsSecondEffect, bool p_bCanBeSecondEffect)
	:	m_eEffectType			( p_eEffectType)
	,	m_bAllowsSecondEffect	( p_bAllowsSecondEffect)
	,	m_bCanBeSecondEffect	( p_bCanBeSecondEffect)
	,	m_iImgWidth				( 0)
	,	m_iImgHeight			( 0)
	,	m_strName				( p_strName)
{
}

ImageEffects::EffectFunctor :: ~EffectFunctor()
{
}

void ImageEffects::EffectFunctor :: SetImage( const wxImage & p_image)
{
	m_iImgWidth = p_image.GetWidth();
	m_iImgHeight = p_image.GetHeight();
}

//*************************************************************************************************

ImageEffects::NoneFunctor :: NoneFunctor()
	:	EffectFunctor( eEffectNone, wxT( "None"), true, true)
{
}

ImageEffects::NoneFunctor :: ~NoneFunctor()
{
}

void ImageEffects::NoneFunctor :: operator ()( const PixelBuffer & p_bufferIn, PixelBuffer & p_bufferOut)
{
	p_bufferOut = p_bufferIn;
}

//*************************************************************************************************

const int ImageEffects::GaussianBlurFunctor :: m_iGaussLvl			= 7;
const int ImageEffects::GaussianBlurFunctor :: m_iGaussMask[]		= {1, 6, 15, 20, 15, 6, 1};
const int ImageEffects::GaussianBlurFunctor :: m_iGaussSum			= 64;

ImageEffects::GaussianBlurFunctor :: GaussianBlurFunctor()
	:	EffectFunctor( eEffectGaussianBlur, wxT( "Gaussian blur"), true, true)
	,	m_temporaryBuffer( Point<size_t, 2>( 100, 100))
{
}

ImageEffects::GaussianBlurFunctor :: ~GaussianBlurFunctor()
{
}

void ImageEffects::GaussianBlurFunctor :: operator ()( const PixelBuffer & p_bufferIn, PixelBuffer & p_bufferOut)
{
	Pixel<double> l_linc;

//	for (int i = m_iGaussLvl - 1 ; i < m_iImgHeight - 1 ; i++)
	for (int i = m_iGaussLvl - 1 ; i < m_iImgHeight ; i++)
	{
//		for (int j = 1 ; j < m_iImgWidth - 1 ; j++)
		for (int j = 0 ; j < m_iImgWidth ; j++)
		{
			l_linc  =	p_bufferIn[i - 6][j];
			l_linc +=	p_bufferIn[i - 5][j].Mul( 6);
			l_linc +=	p_bufferIn[i - 4][j].Mul( 15);
			l_linc +=	p_bufferIn[i - 3][j].Mul( 20);
			l_linc +=	p_bufferIn[i - 2][j].Mul( 15);
			l_linc +=	p_bufferIn[i - 1][j].Mul( 6);
			l_linc +=	p_bufferIn[i - 0][j];
			l_linc /=	m_iGaussSum;

			m_temporaryBuffer[i - 3][j].Set( l_linc.r, l_linc.g, l_linc.b, 255.0);
		}
	}

//	for (int i = 1 ; i < m_iImgHeight - 1 ; i++)
	for (int i = 0 ; i < m_iImgHeight ; i++)
	{
//		for (int j = m_iGaussLvl - 1 ; j < m_iImgWidth - 1 ; j++)
		for (int j = m_iGaussLvl - 1 ; j < m_iImgWidth ; j++)
		{
			l_linc  =	m_temporaryBuffer[i][j - 6];
			l_linc +=	m_temporaryBuffer[i][j - 5].Mul( 6);
			l_linc +=	m_temporaryBuffer[i][j - 4].Mul( 15);
			l_linc +=	m_temporaryBuffer[i][j - 3].Mul( 20);
			l_linc +=	m_temporaryBuffer[i][j - 2].Mul( 15);
			l_linc +=	m_temporaryBuffer[i][j - 1].Mul( 6);
			l_linc +=	m_temporaryBuffer[i][j - 0];
			l_linc /=	m_iGaussSum;

			p_bufferOut[i][j - 3].Set( l_linc.r, l_linc.g, l_linc.b, 255.0);
		}
	}
}

void ImageEffects::GaussianBlurFunctor :: SetImage( const wxImage & p_image)
{
	EffectFunctor::SetImage( p_image);
	m_temporaryBuffer.init( Point<size_t, 2>( m_iImgWidth, m_iImgHeight));
}

//*************************************************************************************************

const double ImageEffects::SharpenFunctor :: m_dSharpenZ		= 1.0;
const double ImageEffects::SharpenFunctor :: m_dSharpenMask[]	= {	0,				-m_dSharpenZ,		0,
																	-m_dSharpenZ,	1+4*m_dSharpenZ,	-m_dSharpenZ,
																	0,				-m_dSharpenZ,		0 };

ImageEffects::SharpenFunctor :: SharpenFunctor()
	:	EffectFunctor( eEffectSharpen, wxT( "Sharpen"), true, true)
{
}

ImageEffects::SharpenFunctor :: ~SharpenFunctor()
{
}

void ImageEffects::SharpenFunctor :: operator ()( const PixelBuffer & p_bufferIn, PixelBuffer & p_bufferOut)
{
	Pixel<double> l_linc;
	UbPixel l_pixel;

	for (int i = 1 ; i < m_iImgHeight - 1 ; i++)
	{
		for (int j = 1 ; j < m_iImgWidth - 1 ; j++)
		{
			l_linc.Set( 0.0, 0.0, 0.0, 255.0);

			l_linc += p_bufferIn[i - 1][j - 1].Mul( m_dSharpenMask[0]);
			l_linc += p_bufferIn[i - 1][j + 0].Mul( m_dSharpenMask[1]);
			l_linc += p_bufferIn[i - 1][j + 1].Mul( m_dSharpenMask[2]);
			l_linc += p_bufferIn[i + 0][j - 1].Mul( m_dSharpenMask[3]);
			l_linc += p_bufferIn[i + 0][j + 0].Mul( m_dSharpenMask[4]);
			l_linc += p_bufferIn[i + 0][j + 1].Mul( m_dSharpenMask[5]);
			l_linc += p_bufferIn[i + 1][j - 1].Mul( m_dSharpenMask[6]);
			l_linc += p_bufferIn[i + 1][j + 0].Mul( m_dSharpenMask[7]);
			l_linc += p_bufferIn[i + 1][j + 1].Mul( m_dSharpenMask[8]);

			p_bufferOut[i][j].Set( clamp( 0.0, 255.0, l_linc.r), clamp( 0.0, 255.0, l_linc.g), clamp( 0.0, 255.0, l_linc.b), 255.0);
		}
	}
}

//*************************************************************************************************

const int ImageEffects::GaussianBlurSharpenFunctor :: m_iGaussLvl			= 7;
const int ImageEffects::GaussianBlurSharpenFunctor :: m_iGaussMask[]		= {1, 6, 15, 20, 15, 6, 1};
const int ImageEffects::GaussianBlurSharpenFunctor :: m_iGaussSum			= 64;
const double ImageEffects::GaussianBlurSharpenFunctor :: m_dSharpenZ		= 1.0;
const double ImageEffects::GaussianBlurSharpenFunctor :: m_dSharpenMask[]	= {	0,				-m_dSharpenZ,		0,
																				-m_dSharpenZ,	1+4*m_dSharpenZ,	-m_dSharpenZ,
																				0,				-m_dSharpenZ,		0 };

ImageEffects::GaussianBlurSharpenFunctor :: GaussianBlurSharpenFunctor()
	:	EffectFunctor( eEffectGaussianBlurSharpen, wxT( "Gaussian blur + sharpen"), true, true)
	,	m_temporaryBuffer1( Point<size_t, 2>( 100, 100))
	,	m_temporaryBuffer2( Point<size_t, 2>( 100, 100))
{
}

ImageEffects::GaussianBlurSharpenFunctor :: ~GaussianBlurSharpenFunctor()
{
}

void ImageEffects::GaussianBlurSharpenFunctor :: operator ()( const PixelBuffer & p_bufferIn, PixelBuffer & p_bufferOut)
{
	Pixel<double> l_linc;
	UbPixel l_pixel;

//	for (int i = m_iGaussLvl - 1 ; i < m_iImgHeight - 1 ; i++)
	for (int i = m_iGaussLvl - 1 ; i < m_iImgHeight ; i++)
	{
//		for (int j = 1 ; j < m_iImgWidth - 1  ; j++)
		for (int j = 0 ; j < m_iImgWidth ; j++)
		{
			l_linc  =	p_bufferIn[i - 6][j];
			l_linc +=	p_bufferIn[i - 5][j].Mul( 6);
			l_linc +=	p_bufferIn[i - 4][j].Mul( 15);
			l_linc +=	p_bufferIn[i - 3][j].Mul( 20);
			l_linc +=	p_bufferIn[i - 2][j].Mul( 15);
			l_linc +=	p_bufferIn[i - 1][j].Mul( 6);
			l_linc +=	p_bufferIn[i - 0][j];
			l_linc /=	m_iGaussSum;

			m_temporaryBuffer1[i - 3][j].Set( l_linc.r, l_linc.g, l_linc.b, 255.0);
		}
	}

//	for (int i = 1 ; i < m_iImgHeight - 1 ; i++)
	for (int i = 0 ; i < m_iImgHeight ; i++)
	{
//		for (int j = m_iGaussLvl - 1 ; j < m_iImgWidth - 1 ; j++)
		for (int j = m_iGaussLvl - 1 ; j < m_iImgWidth ; j++)
		{
			l_linc  =	m_temporaryBuffer1[i][j - 6];
			l_linc +=	m_temporaryBuffer1[i][j - 5].Mul( 6);
			l_linc +=	m_temporaryBuffer1[i][j - 4].Mul( 15);
			l_linc +=	m_temporaryBuffer1[i][j - 3].Mul( 20);
			l_linc +=	m_temporaryBuffer1[i][j - 2].Mul( 15);
			l_linc +=	m_temporaryBuffer1[i][j - 1].Mul( 6);
			l_linc +=	m_temporaryBuffer1[i][j - 0];
			l_linc /=	m_iGaussSum;

			m_temporaryBuffer2[i][j - 3].Set( l_linc.r, l_linc.g, l_linc.b, 255.0);
		}
	}

	for (int i = 1 ; i < m_iImgHeight - 1 ; i++)
	{
		for (int j = 1 ; j < m_iImgWidth - 1 ; j++)
		{
			l_linc.Set( 0.0, 0.0, 0.0, 255.0);

			l_linc += m_temporaryBuffer2[i - 1][j - 1].Mul( m_dSharpenMask[0]);
			l_linc += m_temporaryBuffer2[i - 1][j + 0].Mul( m_dSharpenMask[1]);
			l_linc += m_temporaryBuffer2[i - 1][j + 1].Mul( m_dSharpenMask[2]);
			l_linc += m_temporaryBuffer2[i + 0][j - 1].Mul( m_dSharpenMask[3]);
			l_linc += m_temporaryBuffer2[i + 0][j + 0].Mul( m_dSharpenMask[4]);
			l_linc += m_temporaryBuffer2[i + 0][j + 1].Mul( m_dSharpenMask[5]);
			l_linc += m_temporaryBuffer2[i + 1][j - 1].Mul( m_dSharpenMask[6]);
			l_linc += m_temporaryBuffer2[i + 1][j + 0].Mul( m_dSharpenMask[7]);
			l_linc += m_temporaryBuffer2[i + 1][j + 1].Mul( m_dSharpenMask[8]);

			p_bufferOut[i][j].Set( clamp( 0.0, 255.0, l_linc.r), clamp( 0.0, 255.0, l_linc.g), clamp( 0.0, 255.0, l_linc.b), 255.0);
		}
	}
}

void ImageEffects::GaussianBlurSharpenFunctor :: SetImage( const wxImage & p_image)
{
	EffectFunctor::SetImage( p_image);
	m_temporaryBuffer1.init( Point<size_t, 2>( m_iImgWidth, m_iImgHeight));
	m_temporaryBuffer2.init( Point<size_t, 2>( m_iImgWidth, m_iImgHeight));
}

//*************************************************************************************************

const int ImageEffects::LinearNoiseReductionFunctor :: m_iLnrMask[]	= {	1,	1,	1,
																		1,	1,	1,
																		1,	1,	1};

ImageEffects::LinearNoiseReductionFunctor :: LinearNoiseReductionFunctor()
	:	EffectFunctor( eEffectLinearNoiseReduction, wxT( "Linear noise reduction"), true, true)
{
}

ImageEffects::LinearNoiseReductionFunctor :: ~LinearNoiseReductionFunctor()
{
}

void ImageEffects::LinearNoiseReductionFunctor :: operator ()( const PixelBuffer & p_bufferIn, PixelBuffer & p_bufferOut)
{
	Pixel<double> l_linc;

	for (int i = 1 ; i < m_iImgHeight - 1 ; i++)
	{
		for (int j = 1 ; j < m_iImgWidth - 1 ; j++)
		{
			l_linc  =	p_bufferIn[i - 1][j - 1].Mul( m_iLnrMask[0]);
			l_linc +=	p_bufferIn[i + 0][j - 1].Mul( m_iLnrMask[1]);
			l_linc +=	p_bufferIn[i + 1][j - 1].Mul( m_iLnrMask[2]);
			l_linc +=	p_bufferIn[i - 1][j + 0].Mul( m_iLnrMask[3]);
			l_linc +=	p_bufferIn[i + 0][j + 0].Mul( m_iLnrMask[4]);
			l_linc +=	p_bufferIn[i + 1][j + 0].Mul( m_iLnrMask[5]);
			l_linc +=	p_bufferIn[i - 1][j + 1].Mul( m_iLnrMask[6]);
			l_linc +=	p_bufferIn[i + 0][j + 1].Mul( m_iLnrMask[7]);
			l_linc +=	p_bufferIn[i + 1][j + 1].Mul( m_iLnrMask[8]);
			l_linc /=	9;

			p_bufferOut[i][j].Set( l_linc.r, l_linc.g, l_linc.b, 255.0);
		}
	}
}

//*************************************************************************************************

const int ImageEffects::AnisotropicNoiseReductionFunctor :: m_iAdThreshold = 16;
//const int ImageEffects::AnisotropicNoiseReductionFunctor :: m_iAdMask[]	= {	1,	1,	1,
//																			1,	1,	1,
//																			1,	1,	1};
const int ImageEffects::AnisotropicNoiseReductionFunctor :: m_iAdMask[]	= {	0,	2,	0,
																			2,	1,	2,
																			0,	2,	0};
const int ImageEffects::AnisotropicNoiseReductionFunctor :: m_iAdSum = 9;

ImageEffects::AnisotropicNoiseReductionFunctor :: AnisotropicNoiseReductionFunctor()
	:	EffectFunctor( eEffectAnisotropicNoiseReduction, wxT( "Anisotropic noise reduction"), true, true)
{
}

ImageEffects::AnisotropicNoiseReductionFunctor :: ~AnisotropicNoiseReductionFunctor()
{
}

void ImageEffects::AnisotropicNoiseReductionFunctor :: operator ()( const PixelBuffer & p_bufferIn, PixelBuffer & p_bufferOut)
{
	Pixel<int> l_diff;
	Pixel<int> l_avrg;
	UbPixel l_pixel;

	for (int i = 1 ; i < m_iImgHeight - 1 ; i++)
	{
		for (int j = 1 ; j < m_iImgWidth - 1 ; j++)
		{
			l_pixel =	p_bufferIn[i][j];
			l_avrg  =	p_bufferIn[i - 1][j - 1].Mul( m_iAdMask[0]);
			l_avrg +=	p_bufferIn[i + 0][j - 1].Mul( m_iAdMask[1]);
			l_avrg +=	p_bufferIn[i + 1][j - 1].Mul( m_iAdMask[2]);
			l_avrg +=	p_bufferIn[i - 1][j + 0].Mul( m_iAdMask[3]);
			l_avrg +=	p_bufferIn[i + 0][j + 0].Mul( m_iAdMask[4]);
			l_avrg +=	p_bufferIn[i + 1][j + 0].Mul( m_iAdMask[5]);
			l_avrg +=	p_bufferIn[i - 1][j + 1].Mul( m_iAdMask[6]);
			l_avrg +=	p_bufferIn[i + 0][j + 1].Mul( m_iAdMask[7]);
			l_avrg +=	p_bufferIn[i + 1][j + 1].Mul( m_iAdMask[8]);
			l_avrg /=	m_iAdSum;
			l_diff = l_avrg - l_pixel;

			if (l_diff.r + l_diff.g + l_diff.b > m_iAdThreshold)
			{
				l_avrg = l_pixel;
			}

			p_bufferOut[i][j].Set( l_avrg.r, l_avrg.g, l_avrg.b, 255);
		}
	}
}

//*************************************************************************************************

double	ImageEffects::WaveFunctor :: m_dDegToRad	= M_PI / 180.0;
double	ImageEffects::WaveFunctor :: m_dPiMult2		= M_PI * 2;
double	ImageEffects::WaveFunctor :: m_dCos[360];
double	ImageEffects::WaveFunctor :: m_dSin[360];

ImageEffects::WaveFunctor :: WaveFunctor()
	:	EffectFunctor( eEffectWave, wxT( "Wave"), true, true)
	,	m_dTimeIndexX( 0.0)
	,	m_dTimeIndexY( 0.0)
	,	m_iDecalX( 0)
	,	m_iDecalY( 0)
	,	m_dTimeScaleX( 1.5)
	,	m_dTimeScaleY( 1.0)
	,	m_dScaleX( 5.0)
	,	m_dScaleY( 5.0)
	,	m_dTranslateX( 1.0)
	,	m_dTranslateY( 0.2)
	,	m_dCurTranslateX( 0.0)
	,	m_dCurTranslateY( 0.0)
	,	m_dNbPixelsX( 5.0)
	,	m_dNbPixelsY( 5.0)
{
	for (int i = 0 ; i < 360 ; i++)
	{
		m_dCos[i] = cos( double( i) * m_dDegToRad);
		m_dSin[i] = sin( double( i) * m_dDegToRad);
	}
}

ImageEffects::WaveFunctor :: ~WaveFunctor()
{
}

void ImageEffects::WaveFunctor :: operator ()( const PixelBuffer & p_bufferIn, PixelBuffer & p_bufferOut)
{
	int x;
	int y;
	UbPixel l_pixel;

	for (int i = 0 ; i < m_iImgHeight ; i++)
	{
		m_iDecalY = int( m_dCurTranslateY + m_dSin[int( m_dTimeIndexY + m_dScaleY * 360.0 * double( i) / double( m_iImgHeight)) % 360] * m_dNbPixelsY);

		if (m_iDecalY >= 0)
		{
			if (i + m_iDecalY < m_iImgHeight)
			{
				y = i + m_iDecalY;
			}
			else
			{
				y = i + m_iDecalY - m_iImgHeight;
			}
		}
		else
		{
			if (i + m_iDecalY >= 0)
			{
				y = i + m_iDecalY;
			}
			else
			{
				y = i + m_iDecalY + m_iImgHeight;
			}
		}

		for (int j = 0 ; j < m_iImgWidth ; j++)
		{
			m_iDecalX = int( m_dCurTranslateX + m_dCos[int( m_dTimeIndexX + m_dScaleX * 360.0 * double( j) / double( m_iImgWidth)) % 360] * m_dNbPixelsX);

			if (m_iDecalX >= 0)
			{
				if (j + m_iDecalX < m_iImgWidth)
				{
					x = j + m_iDecalX;
				}
				else
				{
					x = j + m_iDecalX - m_iImgWidth;
				}
			}
			else
			{
				if (j + m_iDecalX >= 0)
				{
					x = j + m_iDecalX;
				}
				else
				{
					x = j + m_iDecalX + m_iImgWidth;
				}
			}

			p_bufferOut[i][j].Set( p_bufferIn[y][x]);
		}
	}

	m_dTimeIndexX += m_dTimeScaleX * 10.0;
	m_dTimeIndexY += m_dTimeScaleY * 10.0;
	m_dCurTranslateX += m_dTranslateX;
	m_dCurTranslateY += m_dTranslateY;

	if (m_dTimeIndexX >= 360.0)
	{
		m_dTimeIndexX -= 360.0;
	}

	if (m_dTimeIndexY >= 360.0)
	{
		m_dTimeIndexY -= 360.0;
	}
	if (m_dCurTranslateX + m_dNbPixelsX >= m_iImgWidth)
	{
		m_dCurTranslateX -= m_iImgWidth;
	}

	if (m_dCurTranslateY + m_dNbPixelsY >= m_iImgHeight)
	{
		m_dCurTranslateY -= m_iImgHeight;
	}
}

//*************************************************************************************************

ImageEffects::GrayScaleFunctor :: GrayScaleFunctor()
	:	EffectFunctor( eEffectGrayScale, wxT( "Gray scale"), true, true)
{
}

ImageEffects::GrayScaleFunctor :: ~GrayScaleFunctor()
{
}

void ImageEffects::GrayScaleFunctor :: operator ()( const PixelBuffer & p_bufferIn, PixelBuffer & p_bufferOut)
{
	byte l_ucFactor;

	for (int i = 0 ; i < m_iImgHeight ; i++)
	{
		for (int j = 0 ; j < m_iImgWidth ; j++)
		{
			l_ucFactor = byte( double( p_bufferIn[i][j].r) * 0.21 + double( p_bufferIn[i][j].g) * 0.71 + double( p_bufferIn[i][j].b) * 0.07);
			l_ucFactor = std::min( l_ucFactor, byte( 255));
			p_bufferOut[i][j].Set( l_ucFactor, l_ucFactor, l_ucFactor, byte( 255));
			p_bufferOut[i][j].a = 255;
		}
	}
}

//*************************************************************************************************

ImageEffects::SepiaFunctor :: SepiaFunctor()
	:	EffectFunctor( eEffectSepia, wxT( "Sepia"), true, true)
{
}

ImageEffects::SepiaFunctor :: ~SepiaFunctor()
{
}

void ImageEffects::SepiaFunctor :: operator ()( const PixelBuffer & p_bufferIn, PixelBuffer & p_bufferOut)
{
	UbPixel l_pixel;

	for (int i = 0 ; i < m_iImgHeight ; i++)
	{
		for (int j = 0 ; j < m_iImgWidth ; j++)
		{
			l_pixel.r = byte( std::min( 255.0, (p_bufferIn[i][j].r * 0.393) + (p_bufferIn[i][j].g * 0.769) + (p_bufferIn[i][j].b * 0.189)));
			l_pixel.g = byte( std::min( 255.0, (p_bufferIn[i][j].r * 0.349) + (p_bufferIn[i][j].g * 0.686) + (p_bufferIn[i][j].b * 0.168)));
			l_pixel.b = byte( std::min( 255.0, (p_bufferIn[i][j].r * 0.272) + (p_bufferIn[i][j].g * 0.534) + (p_bufferIn[i][j].b * 0.131)));
			l_pixel.a = 255;

			p_bufferOut[i][j].Set( l_pixel);
		}
	}
}

//*************************************************************************************************

double length( wxPoint p_ptPoint)
{
	return sqrt( double( p_ptPoint.x * p_ptPoint.x + p_ptPoint.y * p_ptPoint.y));
}

ImageEffects::ImplosionFunctor :: ImplosionFunctor()
	:	EffectFunctor( eEffectImplosion, wxT( "Implosion"), true, true)
{
}

ImageEffects::ImplosionFunctor :: ~ImplosionFunctor()
{
}

void ImageEffects::ImplosionFunctor :: operator ()( const PixelBuffer & p_bufferIn, PixelBuffer & p_bufferOut)
{
	Point<double, 2> l_ptCen;
	Point<double, 2> l_ptMCen;
	Point<int, 2> l_ptCoord;
	Point<int, 2> l_ptSize( m_iImgHeight, m_iImgWidth);

	for (int i = 0 ; i < m_iImgHeight ; i++)
	{
		l_ptCen = Point<double, 2>( 0.5 - double( i) / double( m_iImgHeight), 0.0);

		for (int j = 0 ; j < m_iImgWidth ; j++)
		{
			l_ptCen = Point<double, 2>( l_ptCen[0], 0.5 - double( j) / double( m_iImgWidth));
			l_ptMCen = l_ptCen.GetNormalised() * (0.07 * log( l_ptCen.GetLength()));
			l_ptCoord = l_ptMCen * l_ptSize;
			l_ptCoord[0] = std::max( 0, std::min( m_iImgHeight - 1, i + l_ptCoord[0]));
			l_ptCoord[1] = std::max( 0, std::min( m_iImgWidth - 1, j + l_ptCoord[1]));

			p_bufferOut[i][j].Set( p_bufferIn[l_ptCoord[0]][l_ptCoord[1]]);
		}
	}
}

//*************************************************************************************************

ImageEffects::ExplosionFunctor :: ExplosionFunctor()
	:	EffectFunctor( eEffectExplosion, wxT( "Explosion"), true, true)
{
}

ImageEffects::ExplosionFunctor :: ~ExplosionFunctor()
{
}

void ImageEffects::ExplosionFunctor :: operator ()( const PixelBuffer & p_bufferIn, PixelBuffer & p_bufferOut)
{
	Point<double, 2> l_ptCen;
	Point<double, 2> l_ptMCen;
	Point<int, 2> l_ptCoord;
	Point<int, 2> l_ptSize( m_iImgHeight, m_iImgWidth);

	for (int i = 0 ; i < m_iImgHeight ; i++)
	{
		l_ptCen = Point<double, 2>( 0.5 - double( i) / double( m_iImgHeight), 0.0);

		for (int j = 0 ; j < m_iImgWidth ; j++)
		{
			l_ptCen = Point<double, 2>( l_ptCen[0], 0.5 - double( j) / double( m_iImgWidth));
			l_ptMCen = l_ptCen.GetNormalised() * (-0.07 * log( l_ptCen.GetLength()));
			l_ptCoord = l_ptMCen * l_ptSize;
			l_ptCoord[0] = std::max( 0, std::min( m_iImgHeight - 1, i + l_ptCoord[0]));
			l_ptCoord[1] = std::max( 0, std::min( m_iImgWidth - 1, j + l_ptCoord[1]));

			p_bufferOut[i][j].Set( p_bufferIn[l_ptCoord[0]][l_ptCoord[1]]);
		}
	}
}

//*************************************************************************************************

ImageEffects::WatershedFunctor :: WatershedFunctor()
	:	EffectFunctor( eEffectWatershed, wxT( "Watershed"), true, true)
	,	m_iNbSteps( 1)
{
}

ImageEffects::WatershedFunctor :: ~WatershedFunctor()
{
}

void ImageEffects::WatershedFunctor :: operator ()( const PixelBuffer & p_bufferIn, PixelBuffer & p_bufferOut)
{
	int l_iSteps = (256 / m_iNbSteps);

	for (int i = 0 ; i < m_iImgHeight ; i++)
	{
		for (int j = 0 ; j < m_iImgWidth ; j++)
		{
			p_bufferOut[i][j].Set( (p_bufferIn[i][j] / l_iSteps) * l_iSteps);
		}
	}

	m_iNbSteps++;

	if (m_iNbSteps >= 50)
	{
		m_iNbSteps = 1;
	}
}

//*************************************************************************************************

ImageEffects::RecursiveGaussianBlurFunctor :: RecursiveGaussianBlurFunctor()
	:	EffectFunctor( eEffectRecursiveGaussianBlur, wxT( "Recursive Gaussian blur"), false, false)
{
}

ImageEffects::RecursiveGaussianBlurFunctor :: ~RecursiveGaussianBlurFunctor()
{
}

void ImageEffects::RecursiveGaussianBlurFunctor :: operator ()( const PixelBuffer & p_bufferIn, PixelBuffer & p_bufferOut)
{
	m_onceFunctor( p_bufferIn, p_bufferOut);
}

void ImageEffects::RecursiveGaussianBlurFunctor :: SetImage( const wxImage & p_image)
{
	EffectFunctor::SetImage( p_image);
	m_onceFunctor.SetImage( p_image);
}

//*************************************************************************************************

ImageEffects::RecursiveSharpenFunctor :: RecursiveSharpenFunctor()
	:	EffectFunctor( eEffectRecursiveSharpen, wxT( "Recursive sharpen"), false, false)
{
}

ImageEffects::RecursiveSharpenFunctor :: ~RecursiveSharpenFunctor()
{
}

void ImageEffects::RecursiveSharpenFunctor :: operator ()( const PixelBuffer & p_bufferIn, PixelBuffer & p_bufferOut)
{
	m_onceFunctor( p_bufferIn, p_bufferOut);
}

void ImageEffects::RecursiveSharpenFunctor :: SetImage( const wxImage & p_image)
{
	EffectFunctor::SetImage( p_image);
	m_onceFunctor.SetImage( p_image);
}

//*************************************************************************************************

ImageEffects::RecursiveGaussianBlurSharpenFunctor :: RecursiveGaussianBlurSharpenFunctor()
	:	EffectFunctor( eEffectRecursiveGaussianBlurSharpen, wxT( "Recursive Gaussian blur + sharpen"), false, false)
{
}

ImageEffects::RecursiveGaussianBlurSharpenFunctor :: ~RecursiveGaussianBlurSharpenFunctor()
{
}

void ImageEffects::RecursiveGaussianBlurSharpenFunctor :: operator ()( const PixelBuffer & p_bufferIn, PixelBuffer & p_bufferOut)
{
	m_onceFunctor( p_bufferIn, p_bufferOut);
}

void ImageEffects::RecursiveGaussianBlurSharpenFunctor :: SetImage( const wxImage & p_image)
{
	EffectFunctor::SetImage( p_image);
	m_onceFunctor.SetImage( p_image);
}

//*************************************************************************************************

ImageEffects::RecursiveLinearNoiseReductionFunctor :: RecursiveLinearNoiseReductionFunctor()
	:	EffectFunctor( eEffectRecursiveLinearNoiseReduction, wxT( "Recursive linear noise reduction"), false, false)
{
}

ImageEffects::RecursiveLinearNoiseReductionFunctor :: ~RecursiveLinearNoiseReductionFunctor()
{
}

void ImageEffects::RecursiveLinearNoiseReductionFunctor :: operator ()( const PixelBuffer & p_bufferIn, PixelBuffer & p_bufferOut)
{
	m_onceFunctor( p_bufferIn, p_bufferOut);
}

void ImageEffects::RecursiveLinearNoiseReductionFunctor :: SetImage( const wxImage & p_image)
{
	EffectFunctor::SetImage( p_image);
	m_onceFunctor.SetImage( p_image);
}

//*************************************************************************************************

ImageEffects::RecursiveAnisotropicNoiseReductionFunctor :: RecursiveAnisotropicNoiseReductionFunctor()
	:	EffectFunctor( eEffectRecursiveAnisotropicNoiseReduction, wxT( "Recursive anisotropic reduction"), false, false)
{
}

ImageEffects::RecursiveAnisotropicNoiseReductionFunctor :: ~RecursiveAnisotropicNoiseReductionFunctor()
{
}

void ImageEffects::RecursiveAnisotropicNoiseReductionFunctor :: operator ()( const PixelBuffer & p_bufferIn, PixelBuffer & p_bufferOut)
{
	m_onceFunctor( p_bufferIn, p_bufferOut);
}

void ImageEffects::RecursiveAnisotropicNoiseReductionFunctor :: SetImage( const wxImage & p_image)
{
	EffectFunctor::SetImage( p_image);
	m_onceFunctor.SetImage( p_image);
}

//*************************************************************************************************

ImageEffects::EffectFunctor * ImageEffects::Thread::m_pFunctors[ImageEffects::eEffectsCount];

ImageEffects::Thread :: Thread( ProceduralGenerator * p_pParent, size_t p_uiIndex, int iWidth, int iTop, int iBottom, int iTotalHeight, const UbPixel & p_pxColour)
	:	ProceduralGenerator::Thread( p_pParent, p_uiIndex, iWidth, iTop, iBottom, iTotalHeight, p_pxColour)
	,	m_firstInBuffer( Point<size_t, 2>( iWidth, iTotalHeight))
	,	m_secondInBuffer( Point<size_t, 2>( iWidth, iTotalHeight))
	,	m_pFirstFunction( m_pFunctors[eEffectNone])
	,	m_pNextFirstFunction( m_pFunctors[eEffectNone])
	,	m_pSecondFunction( m_pFunctors[eEffectNone])
	,	m_pNextSecondFunction( m_pFunctors[eEffectNone])
	,	m_pFirstFunctionBufferIn( & m_firstInBuffer)
{
}

ImageEffects::Thread :: ~Thread()
{
}

void ImageEffects::Thread :: Step()
{
	if (m_pNextFirstFunction != m_pFirstFunction)
	{
		( * m_backBuffer) = m_firstInBuffer;
		m_pFirstFunction = m_pNextFirstFunction;
		m_pFirstFunctionBufferIn = m_pNextFirstFunctionBufferIn;
	}
	if (m_pNextSecondFunction != m_pSecondFunction)
	{
		( * m_backBuffer) = m_firstInBuffer;
		m_pSecondFunction = m_pNextSecondFunction;
	}

	(*m_pFirstFunction)( * m_pFirstFunctionBufferIn, m_secondInBuffer);
	(*m_pSecondFunction)( m_secondInBuffer, * m_backBuffer);
}

void ImageEffects::Thread :: SetImage( const wxImage & p_image)
{
	m_iImgWidth = p_image.GetWidth();
	m_iImgHeight = p_image.GetHeight();

	m_firstInBuffer.init( Point<size_t, 2>( m_iImgWidth, m_iImgHeight));
	m_secondInBuffer.init( Point<size_t, 2>( m_iImgWidth, m_iImgHeight));
	m_firstInBuffer = * m_backBuffer;

	for (int i = eEffectNone ; i < eEffectsCount ; i++)
	{
		m_pFunctors[i]->SetImage( p_image);
	}
}

void ImageEffects::Thread :: SetFirstFunction( eEFFECT_TYPE p_eType)
{
	m_pNextFirstFunction = m_pFunctors[p_eType];

	if ( ! m_pFunctors[p_eType]->IsAllowsSecondEffect())
	{
		m_pNextSecondFunction = m_pFunctors[eEffectNone];
		m_pNextFirstFunctionBufferIn = m_backBuffer;
	}
	else
	{
		m_pNextFirstFunctionBufferIn = & m_firstInBuffer;
	}
}

void ImageEffects::Thread :: SetSecondFunction( eEFFECT_TYPE p_eType)
{
	if (m_pFirstFunction->IsAllowsSecondEffect() && m_pFunctors[p_eType]->IsCanBeSecondEffect())
	{
		m_pNextSecondFunction = m_pFunctors[p_eType];
	}
}

void ImageEffects::Thread :: InitialiseFunctors( ImageEffects * p_pParent)
{
	for (int i = eEffectNone ; i < eEffectsCount ; i++)
	{
		switch (i)
		{
		case eEffectNone:
			m_pFunctors[i] = new NoneFunctor;
			break;

		case eEffectGaussianBlur:
			m_pFunctors[i] = new GaussianBlurFunctor;
			break;

		case eEffectSharpen:
			m_pFunctors[i] = new SharpenFunctor;
			break;

		case eEffectGaussianBlurSharpen:
			m_pFunctors[i] = new GaussianBlurSharpenFunctor;
			break;

		case eEffectLinearNoiseReduction:
			m_pFunctors[i] = new LinearNoiseReductionFunctor;
			break;

		case eEffectAnisotropicNoiseReduction:
			m_pFunctors[i] = new AnisotropicNoiseReductionFunctor;
			break;

		case eEffectWave:
			m_pFunctors[i] = new WaveFunctor;
			break;

		case eEffectGrayScale:
			m_pFunctors[i] = new GrayScaleFunctor;
			break;

		case eEffectSepia:
			m_pFunctors[i] = new SepiaFunctor;
			break;

		case eEffectImplosion:
			m_pFunctors[i] = new ImplosionFunctor;
			break;

		case eEffectExplosion:
			m_pFunctors[i] = new ExplosionFunctor;
			break;

		case eEffectWatershed:
			m_pFunctors[i] = new WatershedFunctor;
			break;

		case eEffectRecursiveGaussianBlur:
			m_pFunctors[i] = new RecursiveGaussianBlurFunctor;
			break;

		case eEffectRecursiveSharpen:
			m_pFunctors[i] = new RecursiveSharpenFunctor;
			break;

		case eEffectRecursiveGaussianBlurSharpen:
			m_pFunctors[i] = new RecursiveGaussianBlurSharpenFunctor;
			break;

		case eEffectRecursiveLinearNoiseReduction:
			m_pFunctors[i] = new RecursiveLinearNoiseReductionFunctor;
			break;

		case eEffectRecursiveAnisotropicNoiseReduction:
			m_pFunctors[i] = new RecursiveAnisotropicNoiseReductionFunctor;
			break;
		}

		m_pFunctors[i]->SetGenerator( p_pParent);
	}
}

void ImageEffects::Thread :: CleanupFunctors()
{
	for (int i = eEffectNone ; i < eEffectsCount ; i++)
	{
		delete m_pFunctors[i];
		m_pFunctors[i] = NULL;
	}
}

//*************************************************************************************************

ImageEffects :: ImageEffects( int p_width, int p_height, int p_iWndId, wxFrame * p_pFrame)
	:	ProceduralGenerator( p_width, p_height, p_iWndId, p_pFrame)
{
	Thread::InitialiseFunctors( this);
	m_bNeedsConfigPanel = true;
	wxSystemOptions::SetOption( wxT( "msw.remap"), 0);
	wxInitAllImageHandlers();

	ImageEffects::Thread * l_pThread;
	l_pThread = CreateThread<ImageEffects::Thread>( p_width, 0, p_height, p_height, m_pxColour);
	l_pThread->m_backBuffer = & m_backBuffer;
	l_pThread->Run();
}

ImageEffects :: ~ImageEffects()
{
	Thread::CleanupFunctors();
	wxImage::CleanUpHandlers();
	m_initialised = false;

	_cleanup();
}

void ImageEffects :: SetFirstFunction( eEFFECT_TYPE p_eType)
{
	for (size_t i = 0 ; i < m_arraySlaveThreads.size() ; i++)
	{
		if (m_arraySlaveThreads[i] != NULL)
		{
			static_cast<ImageEffects::Thread *>( m_arraySlaveThreads[i])->SetFirstFunction( p_eType);
		}
	}
}

void ImageEffects :: SetSecondFunction( eEFFECT_TYPE p_eType)
{
	for (size_t i = 0 ; i < m_arraySlaveThreads.size() ; i++)
	{
		if (m_arraySlaveThreads[i] != NULL)
		{
			static_cast<ImageEffects::Thread *>( m_arraySlaveThreads[i])->SetSecondFunction( p_eType);
		}
	}
}

void ImageEffects :: SetImagePath( const wxString & p_strImagePath)
{
	m_initialised = false;
	wxSize l_size = _loadImage( p_strImagePath, m_image, m_backBuffer);
	m_frontBuffer.init( Point<size_t, 2>( l_size.x, l_size.y));
	SetSize( l_size);

	for (size_t i = 0 ; i < m_arraySlaveThreads.size() ; i++)
	{
		if (m_arraySlaveThreads[i] != NULL)
		{
			static_cast<ImageEffects::Thread *>( m_arraySlaveThreads[i])->SetImage( m_image);
		}
	}

	m_initialised = true;
}

ImageEffects::EffectFunctor * ImageEffects :: GetFunctor( eEFFECT_TYPE p_eType)
{
	return Thread::m_pFunctors[p_eType];
}

//*************************************************************************************************

IMPLEMENT_PLUGIN( IePlugin, IePanel, ImageEffects)

IePlugin :: IePlugin()
{
}

IePlugin :: ~IePlugin()
{
}

wxString IePlugin :: GetName()
{
	return wxT( "Image Effects");
}

//*************************************************************************************************

IePanel :: IePanel( wxWindow * p_pWindow, wxWindowID p_id, const wxPoint & p_ptPosition, const wxSize & p_size)
	:	ConfigPanel( p_pWindow, p_id, p_ptPosition, p_size)
{
}

IePanel :: ~IePanel()
{
}

void IePanel :: Create( ImageEffects * p_pGenerator, wxWindow * p_pParent, const wxPoint & p_ptPosition, const wxSize & p_size)
{
	SetSize( p_ptPosition.x, p_ptPosition.y, p_size.x, p_size.y);
	Reparent( p_pParent);
	m_pGenerator = p_pGenerator;

	wxArrayString l_arrayStringFirsts;
	wxArrayString l_arrayStringSeconds;
	ImageEffects::EffectFunctor * l_pFunctor;

	for (int i = ImageEffects::eEffectNone ; i < ImageEffects::eEffectsCount ; i++)
	{
		l_pFunctor = ImageEffects::GetFunctor( ImageEffects::eEFFECT_TYPE( i));
		l_arrayStringFirsts.push_back( l_pFunctor->GetName());

		if (l_pFunctor->IsCanBeSecondEffect())
		{
			l_arrayStringSeconds.push_back( l_pFunctor->GetName());
		}
	}

	m_pStaticFirstFunction = new wxStaticText( this, wxID_ANY, wxT( "First function"), wxPoint( 20, 0), wxSize( p_size.x - 40, 20));
	m_pComboFirstFunction = new wxComboBox( this, eFirstFunction, wxT( "None"), wxPoint( 20, 20), wxSize( p_size.x - 40, 25), l_arrayStringFirsts, wxCB_READONLY);
	m_pComboFirstFunction->SetFont( GetFont());

	m_pStaticSecondFunction = new wxStaticText( this, wxID_ANY, wxT( "Second function"), wxPoint( 20, 50), wxSize( p_size.x - 40, 20));
	m_pComboSecondFunction = new wxComboBox( this, eSecondFunction, wxT( "None"), wxPoint( 20, 70), wxSize( p_size.x - 40, 25), l_arrayStringSeconds, wxCB_READONLY);
	m_pComboSecondFunction->SetFont( GetFont());

	m_pButtonImage = new wxButton( this, eImagePath, wxT( "Image"), wxPoint( 20, 100), wxSize( p_size.x - 40, 30));
}

BEGIN_EVENT_TABLE( IePanel, ConfigPanel)
	EVT_COMBOBOX(	eFirstFunction,		IePanel::_onFirstFunction)
	EVT_COMBOBOX(	eSecondFunction,	IePanel::_onSecondFunction)
	EVT_BUTTON(		eImagePath,			IePanel::_onImagePath)
END_EVENT_TABLE()

void IePanel :: _onSize( wxSizeEvent & p_event)
{
	p_event.Skip();
}

void IePanel :: _onFirstFunction( wxCommandEvent & p_event)
{
	int l_iChoice = p_event.GetInt();
	m_pGenerator->SetFirstFunction( ImageEffects::eEFFECT_TYPE( l_iChoice));

	if (l_iChoice >= ImageEffects::eEffectRecursiveGaussianBlur)
	{
		m_pComboSecondFunction->SetSelection( 0);
		m_pComboSecondFunction->Hide();
		m_pStaticSecondFunction->Hide();
	}
	else
	{
		m_pComboSecondFunction->Show();
		m_pStaticSecondFunction->Show();
	}
}

void IePanel :: _onSecondFunction( wxCommandEvent & p_event)
{
	int l_iChoice = p_event.GetInt();
	m_pGenerator->SetSecondFunction( ImageEffects::eEFFECT_TYPE( l_iChoice));
}

void IePanel :: _onImagePath( wxCommandEvent & p_event)
{
	wxString l_strWildcard = wxT( "Toutes Images (*.bmp;*.gif;*.png;*.jpg)|*.bmp;*.gif;*.png;*.jpg|Fichiers BITMAP (*.bmp)|*.bmp|Fichiers GIF (*.gif)|*.gif|Fichiers JPEG (*.jpg)|*.jpg|Fichiers PNG (*.png)|*.png");
	wxFileDialog l_fileDialog( GetParent(), wxT( "Choose an image"), wxEmptyString, wxEmptyString, l_strWildcard);

	if (l_fileDialog.ShowModal() == wxID_OK)
	{
		wxString l_strImagePath = l_fileDialog.GetPath();
		m_pGenerator->SetImagePath( l_strImagePath);
	}
}

//*************************************************************************************************
