#include "LibGenerator/PrecompiledHeader.h"

#include "LibGenerator/ProceduralGenerator.h"
#include "LibGenerator/GlObjects.h"

#ifdef __WXMSW__
#	include<wx/msw/msvcrt.h>
#endif

using namespace ProceduralTextures;

//*************************************************************************************************

#ifdef _WIN32
#	define LibError() GetLastError()
#else
#	define LibError() dlerror()
#endif

//*************************************************************************************************

ProceduralGenerator::Thread :: Thread( ProceduralGenerator * p_pParent, size_t p_uiIndex, int iWidth, int iTop, int iBottom, int iTotalHeight, const UbPixel & p_pxColour)
	:	m_pParent( p_pParent)
	,	m_uiIndex( p_uiIndex)
	,	m_iWidth( iWidth)
	,	m_iBottom( iBottom)
	,	m_iTop( iTop)
	,	m_iHeight( iTotalHeight)
	,	m_pxColour( p_pxColour)
	,	m_bEnded( true)
	,	m_bLaunched( false)
	,	m_bStopped( false)
{
}

ProceduralGenerator::Thread :: ~Thread()
{
	Kill();
}

void ProceduralGenerator::Thread :: Run()
{
	CreateThread( StEntry, this);
}

void ProceduralGenerator::Thread :: Wait()
{
	Stop();
	ProceduralTextures::Thread::Wait( 1000);
}

unsigned int ProceduralGenerator::Thread :: StEntry( void * p_pThis)
{
	return static_cast<ProceduralGenerator::Thread *>( p_pThis)->Entry();
}

int ProceduralGenerator::Thread :: Entry()
{
	while ( ! IsStopped())
	{
		if (m_bEnded && m_bLaunched && ! IsStopped())
		{
			m_mutex.lock();
			m_bEnded = false;
			Step();
			m_bEnded = true;
			m_bLaunched = false;
			m_mutex.unlock();
		}

		Sleep( 1);
	}

	return 0;
}

//*************************************************************************************************

ProceduralGenerator :: ProceduralGenerator( int p_width, int p_height, int p_iWndId, wxFrame * p_pMainFrame, bool p_bNeedsConfigPanel)
	:	m_sizeDisplay		( p_width, p_height)
	,	m_sizeImage			( p_width, p_height)
	,	m_step				( 0)
	,	m_initialised		( false)
	,	m_pxColour			( 255, 255, 255, 255)
	,	m_bEnded			( true)
	,	m_frontBuffer		( Point<size_t, 2>( p_width, p_height))
	,	m_backBuffer		( Point<size_t, 2>( p_width, p_height))
	,	m_event				( 0, p_iWndId)
	,	m_pMainFrame		( p_pMainFrame)
	,	m_bNeedsConfigPanel	( p_bNeedsConfigPanel)
	,	m_pConfigPanel		( NULL)
	,	m_bToSave			( false)
{
	m_frameBufferSave.Resize( m_sizeImage);
	m_frameBuffer.Resize( m_sizeImage);
	m_texture.Resize( m_sizeImage);
	m_frameBuffer.BindTexture( GL_COLOR_ATTACHMENT0, & m_texture);
	m_frameBufferSave.BindTexture( GL_COLOR_ATTACHMENT0, & m_textureSave);
	m_uiThreadCount = GetCPUCount() * 2;
}

ProceduralGenerator :: ~ProceduralGenerator()
{
	_cleanup();
}

bool ProceduralGenerator :: Step()
{
	bool l_bReturn = false;

	if (m_initialised)
	{
		if (m_arraySlaveThreads.size() > 0)
		{
			if (m_bEnded)
			{
				m_bEnded = false;
				m_event.ResetTime();
				InitialiseStep();
			}
			else if (AllEnded())
			{
				SwapBuffers();
				m_event.ResetTime();
				m_pMainFrame->GetEventHandler()->AddPendingEvent( m_event);
				m_bEnded = true;
				l_bReturn = true;
			}
		}
		else
		{
			m_event.ResetTime();
			m_pMainFrame->GetEventHandler()->AddPendingEvent( m_event);
			l_bReturn = false;
		}
	}

	return l_bReturn;
}

void ProceduralGenerator :: SetRed( unsigned char val)
{
	m_pxColour.r = val;

	for (size_t i = 0 ; i < m_arraySlaveThreads.size() ; i++)
	{
		if (m_arraySlaveThreads[i] != NULL)
		{
			m_arraySlaveThreads[i]->SetRed( val);
		}
	}
}

void ProceduralGenerator :: SetGreen( unsigned char val)
{
	m_pxColour.g = val;

	for (size_t i = 0 ; i < m_arraySlaveThreads.size() ; i++)
	{
		if (m_arraySlaveThreads[i] != NULL)
		{
			m_arraySlaveThreads[i]->SetGreen( val);
		}
	}
}

void ProceduralGenerator :: SetBlue( unsigned char val)
{
	m_pxColour.b = val;

	for (size_t i = 0 ; i < m_arraySlaveThreads.size() ; i++)
	{
		if (m_arraySlaveThreads[i] != NULL)
		{
			m_arraySlaveThreads[i]->SetBlue( val);
		}
	}
}

void ProceduralGenerator :: SwapBuffers()
{
	m_frontBuffer = m_backBuffer;
}

void ProceduralGenerator :: InitialiseStep()
{
	for (size_t i = 0 ; i < m_arraySlaveThreads.size() ; i++)
	{
		if (m_arraySlaveThreads[i] != NULL)
		{
			m_arraySlaveThreads[i]->Relaunch();
		}
	}
}

void ProceduralGenerator :: ClearAllThreads()
{
	for (size_t i = 0 ; i < m_arraySlaveThreads.size() ; i++)
	{
		ProceduralGenerator::Thread * l_pThread = m_arraySlaveThreads[i];
		m_arraySlaveThreads[i] = NULL;

		if (l_pThread != NULL)
		{
			l_pThread->Wait();
			delete l_pThread;
		}
	}

	m_arraySlaveThreads.clear();
}

bool ProceduralGenerator :: AllEnded()
{
	bool l_bReturn = true;

	for (size_t i = 0 ; i < m_arraySlaveThreads.size() && l_bReturn ; i++)
	{
		l_bReturn &= m_arraySlaveThreads[i] == NULL || m_arraySlaveThreads[i]->IsEnded();
	}

	return l_bReturn;
}

void ProceduralGenerator :: Suspend()
{
	m_initialised = false;
	ClearAllThreads();
}

void ProceduralGenerator :: SetSize( int p_iWidth, int p_iHeight)
{
	_resize( wxSize( p_iWidth, p_iHeight));
}

void ProceduralGenerator :: SetSize( const wxSize & p_size)
{
	_resize( p_size);
}

void ProceduralGenerator :: Resize( const wxSize & p_size)
{
	m_sizeDisplay = p_size;
}

void ProceduralGenerator :: GlInit()
{
	OpenGl::Initialise();
	OpenGl::ClearColor( 1.0, 1.0, 1.0, 1.0);
	OpenGl::ShadeModel( GL_FLAT);

	m_vertexBuffer.Initialise();
	m_indexBuffer.Initialise();
	m_texture.Initialise();
	m_frameBuffer.Initialise();
	m_frameBufferSave.Initialise();
	m_textureSave.Initialise();

	_glInit();
}

void ProceduralGenerator :: GlPreRender()
{
	m_texture.Initialise( & m_frontBuffer);
	m_frameBuffer.Initialise( this);
	m_textureSave.Initialise( m_sizeImage);
	m_frameBufferSave.Initialise( this);
	_glPreRender();
}

void ProceduralGenerator :: GlRender( bool p_bChanged)
{
	bool l_bChanged( p_bChanged);
	_glRender( l_bChanged);

	if (m_bToSave)
	{
		OpenGl::Viewport( 0, 0, m_sizeImage.x, m_sizeImage.y);
	}
	else
	{
		OpenGl::Viewport( 0, 0, m_sizeDisplay.x, m_sizeDisplay.y);
	}

	if (m_bToSave)
	{
		m_frameBufferSave.Activate();
	}

	OpenGl::MatrixMode( GL_PROJECTION);
	OpenGl::LoadIdentity();
	OpenGl::Ortho( 0.0, 1.0, 0.0, 1.0, 0.0, 1.0);
	OpenGl::MatrixMode( GL_MODELVIEW);
	OpenGl::LoadIdentity();
	OpenGl::Clear( GL_COLOR_BUFFER_BIT);
	OpenGl::Clear( GL_COLOR_BUFFER_BIT);
	m_texture.Activate();

	if (l_bChanged)
	{
		m_texture.UploadGlImage();
	}

	_subRender();
	m_texture.Deactivate();

	if (m_bToSave)
	{
		_saveFrame();
		m_frameBufferSave.Deactivate();
	}
}

void ProceduralGenerator :: GlPostRender()
{
	_glPostRender();
}

void ProceduralGenerator :: GlCleanup()
{
	m_vertexBuffer.Cleanup();
	m_indexBuffer.Cleanup();
	m_texture.Cleanup();
	m_frameBuffer.Cleanup();
	m_frameBufferSave.Cleanup();
	m_textureSave.Cleanup();
	_glCleanup();
	OpenGl::Cleanup();
}

void ProceduralGenerator :: SaveFrame()
{
	m_bToSave = true;
}

void ProceduralGenerator :: _cleanup()
{
	ClearAllThreads();

	m_frontBuffer.clear();
	m_backBuffer.clear();
}

void ProceduralGenerator :: _resize( const wxSize & p_size)
{
	m_sizeImage = p_size;
	m_sizeDisplay = p_size;
	wxSize l_sizeConfig( 0, 0);

	if (m_pConfigPanel != NULL)
	{
		l_sizeConfig = m_pConfigPanel->GetSize();
	}

	wxDisplay l_display;
	wxSize l_maxSize = l_display.GetClientArea().GetSize() + m_pMainFrame->GetClientSize() - m_pMainFrame->GetSize() - wxSize( l_sizeConfig.x, 0);
	double l_dRatio = double( m_sizeDisplay.x) / double( m_sizeDisplay.y);
	bool l_bChanged = false;

	l_maxSize = wxSize( std::min<int>( l_maxSize.x, l_maxSize.y * l_dRatio),
		std::min<int>( l_maxSize.y, l_maxSize.x / l_dRatio));

	while (m_sizeDisplay.x > l_maxSize.x || m_sizeDisplay.y > l_maxSize.y)
	{
		if (m_sizeDisplay.x > l_maxSize.x)
		{
			m_sizeDisplay.x = l_maxSize.x;
			m_sizeDisplay.y = m_sizeDisplay.x / l_dRatio;
		}
		else if (m_sizeDisplay.y > l_maxSize.y)
		{
			m_sizeDisplay.y = l_maxSize.y;
			m_sizeDisplay.x = m_sizeDisplay.y * l_dRatio;
		}

		l_bChanged = true;
	}

	m_pMainFrame->SetClientSize( m_sizeDisplay + wxSize( l_sizeConfig.x, 0));
	m_texture.Resize( m_sizeImage);
	m_frameBuffer.Resize( m_sizeImage);
}

wxSize ProceduralGenerator :: _loadImage( const wxString & p_strImagePath, wxImage & p_image, PixelBuffer & p_buffer)
{
	wxSize l_sizeReturn;

	if (p_image.LoadFile( p_strImagePath))
	{
		wxDisplay l_display;
		wxSize l_sizeConfig;

		if (m_pConfigPanel != NULL)
		{
			l_sizeConfig = m_pConfigPanel->GetSize();
		}

		l_sizeReturn.x = p_image.GetWidth();
		l_sizeReturn.y = p_image.GetHeight();

		p_image = p_image.Mirror( false);
		p_buffer.init( Point<size_t, 2>( l_sizeReturn.x, l_sizeReturn.y));

		if (p_image.HasAlpha())
		{
			p_buffer.set<unsigned char, 3, 1>( p_image.GetData(), p_image.GetAlpha());
		}
		else
		{
			p_buffer.set<unsigned char, 3, 0>( p_image.GetData(), NULL);
		}
	}
	else
	{
		wxMessageBox( wxT( "No handler found for this image format"));
	}

	return l_sizeReturn;
}

void ProceduralGenerator :: _subRender()
{
	if (m_vertexBuffer.Activate())
	{
		if (m_indexBuffer.Activate())
		{
			OpenGl::DrawElements( GL_TRIANGLES, m_indexBuffer.GetSize(), GL_UNSIGNED_INT, 0);
			m_indexBuffer.Deactivate();
		}

		m_vertexBuffer.Deactivate();
	}
}

void ProceduralGenerator :: _saveFrame()
{
	if (m_bToSave)
	{
		wxImage l_image( m_sizeImage.x, m_sizeImage.y, true);
		int x = 0;
		int y;
		m_frameBufferSave.DownloadBuffer( GL_COLOR_ATTACHMENT0, m_frontBuffer);

		for (int i = 0 ; i < m_sizeImage.x ; i++)
		{
			y = m_sizeImage.y - 1;
			for (int j = 0 ; j < m_sizeImage.y ; j++)
			{
				const UbPixel & l_pixel = m_frontBuffer[j][i];
				l_image.SetRGB( x, y, l_pixel.r, l_pixel.g, l_pixel.b);
				y--;
			}
			x++;
		}

		l_image.SaveFile( wxT( "./Image.bmp"), wxBITMAP_TYPE_BMP);
		m_bToSave = false;
	}
}

//*************************************************************************************************

ConfigPanel :: ConfigPanel( wxWindow * p_pWindow, wxWindowID p_id, const wxPoint & p_ptPosition, const wxSize & p_size)
    :   wxPanel( p_pWindow, p_id, p_ptPosition, p_size)
{
	SetBackgroundColour( p_pWindow->GetBackgroundColour());
	SetForegroundColour( p_pWindow->GetForegroundColour());
}

ConfigPanel :: ~ConfigPanel()
{
}

//*************************************************************************************************
