#include "ShaderEffects/GeneratorShaderEffects.h"

#include <wx/sysopt.h>
#include <wx/display.h>
#include <wx/tglbtn.h>
#include <cmath>
#include <wx/timer.h>

#ifdef __WXMSW__
#	include<wx/msw/msvcrt.h>
#endif

using namespace ProceduralTextures;

//*************************************************************************************************

ShaderEffects::Effect :: Effect( wxImage * p_pImage)
	:	m_pImage			( p_pImage)
	,	m_pShaderProgram	( NULL)
	,	m_strVertexFile		()
	,	m_strFragmentFile	()
	,	m_bNewShader		( false)
	,	m_llStartTime		( 0)
	,	m_iWidth			( p_pImage->GetWidth())
	,	m_iHeight			( p_pImage->GetHeight())
{
}

ShaderEffects::Effect :: ~Effect()
{
	delete m_pShaderProgram;

	for (size_t i = 0 ; i < m_arrayImages.size() ; i++)
	{
		delete m_arrayImages[i];
	}

	m_arrayImages.clear();
	m_arrayImagesFiles.clear();
}

void ShaderEffects::Effect :: SetVertexFile( const wxString & p_strPath)
{
	m_strVertexFile = p_strPath;
}

void ShaderEffects::Effect :: SetFragmentFile( const wxString & p_strPath)
{
	m_strFragmentFile = p_strPath;
}

void ShaderEffects::Effect :: Compile()
{
	m_bNewShader = true;
}

void ShaderEffects::Effect :: SetImagePath( size_t p_iImage, const wxString & p_strImagePath)
{
	wxImage * l_pImage = NULL;

	if (p_iImage >= m_arrayImages.size())
	{
		l_pImage = new wxImage();
		m_arrayImages.push_back( l_pImage);
		m_arrayImagesFiles.push_back( p_strImagePath);
	}
	else
	{
		l_pImage = m_arrayImages[p_iImage];
		m_arrayImagesFiles[p_iImage] = p_strImagePath;
	}

	if (l_pImage->LoadFile( p_strImagePath))
	{
		int l_iWidth = l_pImage->GetWidth();
		int l_iHeight = l_pImage->GetHeight();
		std::cout << "Original : " << l_iWidth << "x" << l_iHeight << "\n";

		if (m_iWidth != l_iWidth || m_iHeight != l_iHeight)
		{
			std::cout << "Computed : " << m_iWidth << "x" << m_iHeight << "\n";
			l_pImage->Rescale( m_iWidth, m_iHeight, wxIMAGE_QUALITY_HIGH);
			std::cout << "Scaled : " << l_pImage->GetWidth() << "x" << l_pImage->GetHeight() << "\n";
		}
	}
	else
	{
		wxMessageBox( wxT( "No handler found for this image format"));
	}
}

void ShaderEffects::Effect :: GlPreRender()
{
	if (m_bNewShader)
	{
		if (m_pShaderProgram != NULL)
		{
			m_pShaderProgram->Cleanup();
		}
		else
		{
			m_pShaderProgram = new GlShaderProgram;
		}

		if ( ! m_strVertexFile.empty())
		{
			m_pShaderProgram->CreateObject( eSHADER_OBJECT_TYPE_VERTEX);
			m_pShaderProgram->SetProgramFile( m_strVertexFile, eSHADER_OBJECT_TYPE_VERTEX);
		}

		if ( ! m_strFragmentFile.empty())
		{
			m_pShaderProgram->CreateObject( eSHADER_OBJECT_TYPE_PIXEL);
			m_pShaderProgram->SetProgramFile( m_strFragmentFile, eSHADER_OBJECT_TYPE_PIXEL);
		}

		m_pShaderProgram->Initialise();

		m_pShaderProgram->CreateIntFrameVariable( wxT( "pg_width"),			0);
		m_pShaderProgram->CreateIntFrameVariable( wxT( "pg_height"),		0);
		m_pShaderProgram->CreateIntFrameVariable( wxT( "pg_time"),			0);
		m_pShaderProgram->CreateIntFrameVariable( wxT( "pg_texture"),		0);
		m_pShaderProgram->CreateIntFrameVariable( wxT( "pg_modif1"),		1);
		m_pShaderProgram->CreateIntFrameVariable( wxT( "pg_modif2"),		2);
		m_pShaderProgram->CreateIntFrameVariable( wxT( "pg_modif3"),		3);
		m_pShaderProgram->CreateIntFrameVariable( wxT( "pg_sep_type"),		0);
		m_pShaderProgram->CreateIntFrameVariable( wxT( "pg_sep_offset"),	0);

		m_llStartTime = wxGetLocalTimeMillis();
		m_bNewShader = false;
	}
}

void ShaderEffects::Effect :: GlPostRender()
{
}

void ShaderEffects::Effect :: GlCleanup()
{
	if (m_pShaderProgram != NULL)
	{
		m_pShaderProgram->Cleanup();
	}
}

bool ShaderEffects::Effect :: Activate( eSEPARATION p_eSeparationType, int p_iSeparationOffset)
{
	bool l_bReturn = true;

	if (m_pShaderProgram != NULL)
	{
		m_pShaderProgram->SetIntFrameVariableValue( wxT( "pg_width"),		m_iWidth);
		m_pShaderProgram->SetIntFrameVariableValue( wxT( "pg_height"),		m_iHeight);
		m_pShaderProgram->SetIntFrameVariableValue( wxT( "pg_sep_type"),	p_eSeparationType);
		m_pShaderProgram->SetIntFrameVariableValue( wxT( "pg_sep_offset"),	p_iSeparationOffset);
		m_pShaderProgram->SetIntFrameVariableValue( wxT( "pg_time"),		( wxGetLocalTimeMillis() - m_llStartTime).ToLong());
		m_pShaderProgram->Activate();
	}

	return l_bReturn;
}

void ShaderEffects::Effect :: Deactivate()
{
	if (m_pShaderProgram != NULL)
	{
		m_pShaderProgram->Deactivate();
	}
}

wxString ShaderEffects::Effect :: GetCompilerLog( eSHADER_OBJECT_TYPE p_eType)
{
	wxString l_strReturn;

	if (m_pShaderProgram != NULL)
	{
		GlShaderObject * l_pShader = m_pShaderProgram->GetProgram( p_eType);

		if (l_pShader != NULL)
		{
			l_strReturn = l_pShader->GetCompilerLog();
		}
	}

	return l_strReturn;
}

wxString ShaderEffects::Effect :: GetLinkerLog()
{
	wxString l_strReturn;

	if (m_pShaderProgram != NULL)
	{
		l_strReturn = m_pShaderProgram->GetLinkerLog();
	}

	return l_strReturn;
}

void ShaderEffects::Effect :: UpdateImages()
{
	m_iWidth = m_pImage->GetWidth();
	m_iHeight = m_pImage->GetHeight();

	for (size_t i = 0 ; i < m_arrayImages.size() ; i++)
	{
		m_arrayImages[i]->LoadFile( m_arrayImagesFiles[i]);
		m_arrayImages[i]->ResampleBicubic( m_iWidth, m_iHeight);
	}
}

//*************************************************************************************************

ShaderEffects :: ShaderEffects( int p_width, int p_height, int p_iWndId, wxFrame * p_pFrame)
	:	ProceduralGenerator( p_width, p_height, p_iWndId, p_pFrame, true)
	,	m_iSeparationOffset	( 50)
	,	m_eSeparationType	( eSEPARATION_NONE)
{
	m_texture2.Resize( m_sizeImage);
	m_frameBuffer1.Resize( m_sizeImage);
	m_frameBuffer2.Resize( m_sizeImage);
	m_frameBuffer1.BindTexture( GL_COLOR_ATTACHMENT0, & m_texture1);
	m_frameBuffer2.BindTexture( GL_COLOR_ATTACHMENT0, & m_texture2);
	wxSystemOptions::SetOption( wxT( "msw.remap"), 0);
	wxInitAllImageHandlers();
}

ShaderEffects :: ~ShaderEffects()
{
	for (size_t i = 0 ; i < m_arrayEffects.size() ; i++)
	{
		delete m_arrayEffects[i];
	}

	m_arrayEffects.clear();
	wxImage::CleanUpHandlers();
	m_initialised = false;
	_cleanup();
}

void ShaderEffects :: SetImagePath( const wxString & p_strImagePath)
{
	m_initialised = false;
	wxSize l_size = _loadImage( p_strImagePath, m_image, m_backBuffer);
	m_frontBuffer.init( Point<size_t, 2>( l_size.x, l_size.y));
	SetSize( l_size);

	if (m_arrayEffects.size() == 0)
	{
		EffectAdd();
	}
	else
	{
		for (size_t i = 0 ; i < m_arrayEffects.size() ; i++)
		{
			m_arrayEffects[i]->UpdateImages();
		}
	}

	m_initialised = true;
}

void ShaderEffects :: ResetTime()
{
	for (size_t i = 0 ; i < m_arrayEffects.size() ; i++)
	{
		m_arrayEffects[i]->ResetTime();
	}
}

void ShaderEffects :: _glInit()
{
	m_texture1.Initialise();
	m_texture2.Initialise();

	m_frameBuffer1.Initialise();
	m_frameBuffer2.Initialise();
}

void ShaderEffects :: _glPreRender()
{
	m_texture1.Initialise( & m_backBuffer);
	m_texture2.Initialise( m_sizeImage);

	m_frameBuffer1.Initialise( this);
	m_frameBuffer2.Initialise( this);

	for (size_t i = 0 ; i < m_arrayEffects.size() ; i++)
	{
		m_arrayEffects[i]->GlPreRender();
	}

	OpenGl::Viewport( 0, 0, m_sizeImage.x, m_sizeImage.y);
	OpenGl::MatrixMode( GL_PROJECTION);
	OpenGl::LoadIdentity();
	OpenGl::Ortho( 0.0, 1.0, 0.0, 1.0, 0.0, 1.0);
	OpenGl::MatrixMode( GL_MODELVIEW);
	OpenGl::LoadIdentity();
}

void ShaderEffects :: _glRender( bool & p_bChanged)
{
	GlFrameBuffer * l_pFrameBufferTexture = & m_frameBuffer1;
	GlFrameBuffer * l_pFrameBufferRender = & m_frameBuffer2;

	m_texture1.Activate();
	m_texture1.UploadGlImage();
	m_texture1.Deactivate();

	for (size_t i = 0 ; i < m_arrayEffects.size() ; i++)
	{
		if (l_pFrameBufferRender->Activate())
		{
			if (EffectActivate( i))
			{
				OpenGl::Clear( GL_COLOR_BUFFER_BIT);
				l_pFrameBufferTexture->GetTexture( GL_COLOR_ATTACHMENT0)->Activate();
				_subRender();
				l_pFrameBufferTexture->GetTexture( GL_COLOR_ATTACHMENT0)->Deactivate();
				EffectDeactivate( i);
			}

			l_pFrameBufferRender->Deactivate();
		}

		std::swap( l_pFrameBufferTexture, l_pFrameBufferRender);
	}

	m_frameBuffer.Activate();
	OpenGl::Clear( GL_COLOR_BUFFER_BIT);
	l_pFrameBufferTexture->GetTexture( GL_COLOR_ATTACHMENT0)->Activate();
	_subRender();
	l_pFrameBufferTexture->GetTexture( GL_COLOR_ATTACHMENT0)->Deactivate();
	m_frameBuffer.Deactivate();
}

void ShaderEffects :: _glPostRender()
{
	for (size_t i = 0 ; i < m_arrayEffects.size() ; i++)
	{
		m_arrayEffects[i]->GlPostRender();
	}
}

void ShaderEffects :: _glCleanup()
{
	for (size_t i = 0 ; i < m_arrayEffects.size() ; i++)
	{
		m_arrayEffects[i]->GlCleanup();
		delete m_arrayEffects[i];
	}

	m_arrayEffects.clear();
}

bool ShaderEffects :: EffectAdd()
{
	m_arrayEffects.push_back( new Effect( & m_image));
	return true;
}

bool ShaderEffects :: EffectRemove( size_t p_uiIndex)
{
	bool l_bReturn = p_uiIndex < m_arrayEffects.size();

	if (l_bReturn)
	{
		Effect * l_pEffect = m_arrayEffects[p_uiIndex];
		std::vector <Effect *>::iterator l_it = m_arrayEffects.begin();
		m_arrayEffects.erase( l_it + p_uiIndex);
		delete l_pEffect;
	}

	return l_bReturn;
}

void ShaderEffects :: EffectSetVertexFile( size_t p_uiIndex, const wxString & p_strPath)
{
	if (p_uiIndex < m_arrayEffects.size())
	{
		m_arrayEffects[p_uiIndex]->SetVertexFile( p_strPath);
	}
}

void ShaderEffects :: EffectSetFragmentFile( size_t p_uiIndex, const wxString & p_strPath)
{
	if (p_uiIndex < m_arrayEffects.size())
	{
		m_arrayEffects[p_uiIndex]->SetFragmentFile( p_strPath);
	}
}

void ShaderEffects :: EffectSetImagePath( size_t p_uiIndex, size_t p_iImage, const wxString & p_strImagePath)
{
	m_initialised = false;

	if (p_uiIndex < m_arrayEffects.size())
	{
		m_arrayEffects[p_uiIndex]->SetImagePath( p_iImage, p_strImagePath);
	}

	m_initialised = true;
}

void ShaderEffects :: EffectCompile( size_t p_uiIndex)
{
	if (p_uiIndex < m_arrayEffects.size())
	{
		m_arrayEffects[p_uiIndex]->Compile();
	}
}

wxString ShaderEffects :: EffectGetCompilerLog( size_t p_uiIndex, eSHADER_OBJECT_TYPE p_eType)
{
	wxString l_strReturn;

	if (p_uiIndex < m_arrayEffects.size())
	{
		m_arrayEffects[p_uiIndex]->GetCompilerLog( p_eType);
	}

	return l_strReturn;
}

wxString ShaderEffects :: EffectGetLinkerLog( size_t p_uiIndex)
{
	wxString l_strReturn;

	if (p_uiIndex < m_arrayEffects.size())
	{
		m_arrayEffects[p_uiIndex]->GetLinkerLog();
	}

	return l_strReturn;
}

bool ShaderEffects :: EffectActivate( size_t p_uiIndex)
{
	bool l_bReturn = p_uiIndex < m_arrayEffects.size();

	if (l_bReturn)
	{
		l_bReturn = m_arrayEffects[p_uiIndex]->Activate( m_eSeparationType, m_iSeparationOffset);
	}

	return l_bReturn;
}

void ShaderEffects :: EffectDeactivate( size_t p_uiIndex)
{
	if (p_uiIndex < m_arrayEffects.size())
	{
		m_arrayEffects[p_uiIndex]->Deactivate();
	}
}

void ShaderEffects :: _resize( const wxSize & p_size)
{
	ProceduralGenerator::_resize( p_size);
	m_frameBuffer1.Resize( m_sizeImage);
	m_frameBuffer2.Resize( m_sizeImage);
	m_texture1.Resize( m_sizeImage);
	m_texture2.Resize( m_sizeImage);
}

//******************************************************************************

IMPLEMENT_PLUGIN( SePlugin, SePanel, ShaderEffects)

SePlugin :: SePlugin()
{
}

SePlugin :: ~SePlugin()
{
}

wxString SePlugin :: GetName()
{
	return wxT( "Shader Effects");
}

//*************************************************************************************************

SePanel :: SePanel( wxWindow * p_pWindow, wxWindowID p_id, const wxPoint & p_ptPosition, const wxSize & p_size)
	:	ConfigPanel( p_pWindow, p_id, p_ptPosition, p_size)
	,	m_pShadersConfigs( NULL)
{
}

SePanel :: ~SePanel()
{
}

void SePanel :: Create( ShaderEffects * p_pGenerator, wxWindow * p_pParent, const wxPoint & p_ptPosition, const wxSize & p_size)
{
	SetSize( p_ptPosition.x, p_ptPosition.y, p_size.x, p_size.y);
	Reparent( p_pParent);
	m_pGenerator = p_pGenerator;

	m_pButtonImage = new wxButton( this, eImagePath, wxT( "Image"), wxPoint( 20, 0), wxSize( p_size.x - 40, 30));

	m_pButtonResetTime = new wxButton( this, eResetTime, wxT( "Reset Time"), wxPoint( 20, 35), wxSize( p_size.x - 40, 30));
	m_pButtonResetTime->Hide();

	wxArrayString l_arrayChoices;
	l_arrayChoices.push_back( wxT( "Not separated"));
	l_arrayChoices.push_back( wxT( "Horizontally"));
	l_arrayChoices.push_back( wxT( "Vertically"));
	m_pComboSeparationType = new wxComboBox( this, eSepType, wxT( "Not separated"), wxPoint( 20, 70), wxSize( p_size.x - 40, 25), l_arrayChoices, wxCB_READONLY);
	m_pComboSeparationType->SetFont( GetFont());
	m_pComboSeparationType->Hide();

	m_pSliderSeparationOffset = new wxSlider( this, eSepOffset, 50, 0, 100, wxPoint( 20, 100), wxSize( 100, 20));
	m_pSliderSeparationOffset->Hide();

	m_pShadersConfigs = new SeEffectsNotebook( this, eShaders, wxPoint( 0, 125), wxSize( p_size.x, p_size.y - 135));
	m_pShadersConfigs->CreatePage( m_pGenerator, wxT( "Shader 1"));
	m_pShadersConfigs->CreatePage( m_pGenerator, wxT( "New ..."));
	m_pShadersConfigs->Layout();
	m_pShadersConfigs->Hide();
}

BEGIN_EVENT_TABLE( SePanel, ConfigPanel)
	EVT_SIZE(							SePanel::_onSize)
	EVT_BUTTON(			eImagePath,		SePanel::_onImagePath)
	EVT_COMBOBOX(		eSepType,		SePanel::_onSepType)
	EVT_SLIDER(			eSepOffset,		SePanel::_onSepOffset)
	EVT_BUTTON(			eResetTime,		SePanel::_onResetTime)
END_EVENT_TABLE()

void SePanel :: _onSize( wxSizeEvent & p_event)
{
	if (m_pShadersConfigs != NULL)
	{
		m_pShadersConfigs->SetSize( p_event.GetSize().x, p_event.GetSize().y - 135);
	}

	p_event.Skip();
}

void SePanel :: _onImagePath( wxCommandEvent & p_event)
{
	wxString l_strWildcard = wxT( "Toutes Images (*.bmp;*.gif;*.png;*.jpg)|*.bmp;*.gif;*.png;*.jpg|Fichiers BITMAP (*.bmp)|*.bmp|Fichiers GIF (*.gif)|*.gif|Fichiers JPEG (*.jpg)|*.jpg|Fichiers PNG (*.png)|*.png");
	wxString l_strImagePath = wxFileSelector( wxT( "Choose an image"), wxEmptyString, wxEmptyString, wxEmptyString, l_strWildcard);

	if ( ! l_strImagePath.empty())
	{
		m_pGenerator->SetImagePath( l_strImagePath);
		m_pButtonResetTime->Show();
		m_pComboSeparationType->Show();
		m_pShadersConfigs->Show();
	}
}

void SePanel :: _onSepType( wxCommandEvent	& p_event)
{
	m_pGenerator->SetSeparationType( ShaderEffects::eSEPARATION( p_event.GetInt()));

	if (p_event.GetInt() != 0)
	{
		m_pSliderSeparationOffset->Show();
	}
	else
	{
		m_pSliderSeparationOffset->Hide();
	}
}

void SePanel :: _onSepOffset( wxCommandEvent & p_event)
{
	m_pGenerator->SetSeparationOffset( p_event.GetInt());
}

void SePanel :: _onResetTime( wxCommandEvent & p_event)
{
	m_pGenerator->ResetTime();
}

//*************************************************************************************************

SeEffectPage :: SeEffectPage( SeEffectsNotebook * p_pParent, ShaderEffects * p_pGenerator, size_t p_uiIndex, bool p_bCanRemove, const wxPoint & p_ptPosition, const wxSize & p_size)
	:	wxNotebookPage( p_pParent, wxID_ANY, p_ptPosition, p_size, wxBORDER_NONE | wxCLIP_CHILDREN)
	,	m_pParent( p_pParent)
	,	m_uiIndex( p_uiIndex)
	,	m_pGenerator( p_pGenerator)
{
	m_pButtonVertexFile		= new wxToggleButton(	this,	eVertexFile,	wxT( "Vertex shader"),		wxPoint( 0, 0),		wxSize( p_size.x, 30));
	m_pButtonFragmentFile	= new wxToggleButton(	this,	eFragmentFile,	wxT( "Fragment shader"),	wxPoint( 0, 35),	wxSize( p_size.x, 30));
	m_pButtonCompile		= new wxButton(			this,	eCompile,		wxT( "Compile"),			wxPoint( 0, 70),	wxSize( p_size.x, 30));
	m_pButtonCompilerLog	= new wxButton(			this,	eCompilerLog,	wxT( "Compile result"),		wxPoint( 0, 105),	wxSize( p_size.x, 30));

	if (p_bCanRemove)
	{
		m_pButtonRemove		= new wxButton(			this,	eRemove,		wxT( "Remove"),				wxPoint( 0, 140),	wxSize( p_size.x, 30));
	}

	m_pButtonCompile->Disable();
	m_pButtonCompilerLog->Disable();
}

SeEffectPage :: ~SeEffectPage()
{
}

BEGIN_EVENT_TABLE( SeEffectPage, wxNotebookPage)
	EVT_TOGGLEBUTTON(	eVertexFile,	SeEffectPage::_onVertexShaderPath)
	EVT_TOGGLEBUTTON(	eFragmentFile,	SeEffectPage::_onFragmentShaderPath)
	EVT_BUTTON(			eCompile,		SeEffectPage::_onShaderCompile)
	EVT_BUTTON(			eCompilerLog,	SeEffectPage::_onCompilerLog)
	EVT_BUTTON(			eRemove,		SeEffectPage::_onRemove)
END_EVENT_TABLE()

void SeEffectPage :: _onVertexShaderPath( wxCommandEvent & p_event)
{
	wxString l_strFile;

	if (p_event.IsChecked())
	{
		wxString l_strWildcard = wxT( "Tous Fichiers (*.*)|*.*");
		l_strFile = wxFileSelector( wxT( "Choose a file"), wxT( "./data"), wxEmptyString, wxEmptyString, l_strWildcard);
	}

	m_pButtonCompile->Enable();
	m_pButtonCompilerLog->Disable();
	m_pGenerator->EffectSetVertexFile( m_uiIndex, l_strFile);
}

void SeEffectPage :: _onFragmentShaderPath( wxCommandEvent & p_event)
{
	wxString l_strFile;

	if (p_event.IsChecked())
	{
		wxString l_strWildcard = wxT( "Tous Fichiers (*.*)|*.*");
		l_strFile = wxFileSelector( wxT( "Choose a file"), wxT( "./data"), wxEmptyString, wxEmptyString, l_strWildcard);
	}

	m_pButtonCompile->Enable();
	m_pButtonCompilerLog->Disable();
	m_pGenerator->EffectSetFragmentFile( m_uiIndex, l_strFile);
}

void SeEffectPage :: _onShaderCompile( wxCommandEvent & p_event)
{
	m_pGenerator->EffectCompile( m_uiIndex);
	m_pButtonCompile->Disable();
	m_pButtonCompilerLog->Enable();
}

void SeEffectPage :: _onCompilerLog( wxCommandEvent & p_event)
{
	wxString l_strVertexLog = m_pGenerator->EffectGetCompilerLog( m_uiIndex, eSHADER_OBJECT_TYPE_VERTEX);
	wxString l_strFragmentLog = m_pGenerator->EffectGetCompilerLog( m_uiIndex, eSHADER_OBJECT_TYPE_PIXEL);
	wxString l_strLinkerLog = m_pGenerator->EffectGetLinkerLog( m_uiIndex);
	wxString l_strText;

	if ( ! l_strVertexLog.empty())
	{
		l_strText += wxT( "Vertex Shader :\n") + l_strVertexLog;
	}

	if ( ! l_strFragmentLog.empty())
	{
		if ( ! l_strText.empty())
		{
			l_strText += wxT( "\n\n");
		}

		l_strText += wxT( "Fragment Shader :\n") + l_strFragmentLog + wxT( "\n\n");
	}

	if ( ! l_strLinkerLog.empty())
	{
		if ( ! l_strText.empty())
		{
			l_strText += wxT( "\n\n");
		}

		l_strText += wxT( "Linker :\n") + l_strLinkerLog + wxT( "\n\n");
	}

	if (l_strText.empty())
	{
		l_strText = wxT( "No error");
	}

	wxMessageBox( l_strText, wxT( "Compile and Link result"));
}

void SeEffectPage :: _onRemove( wxCommandEvent & p_event)
{
	if (m_pGenerator->EffectRemove( m_uiIndex))
	{
		m_pParent->ChangeSelection( m_uiIndex - 1);
		m_pParent->DeletePage( m_uiIndex);
	}
}

//*************************************************************************************************

SeEffectsNotebook :: SeEffectsNotebook( wxWindow * p_pParent, wxWindowID p_id, const wxPoint & p_ptPosition, const wxSize & p_size, long lStyle)
	:	wxNotebook( p_pParent, p_id, p_ptPosition, p_size, lStyle)
	,	m_pSelection( NULL)
	,	m_pGenerator( NULL)
{
	SetBackgroundColour( p_pParent->GetBackgroundColour());
	SetForegroundColour( p_pParent->GetForegroundColour());
}

SeEffectsNotebook :: ~SeEffectsNotebook()
{
}

SeEffectPage * SeEffectsNotebook :: CreatePage( ShaderEffects * p_pGenerator, const wxString & p_strTitle)
{
	wxSize l_size = GetClientSize();
	SeEffectPage * l_pReturn = new SeEffectPage( this, p_pGenerator, m_listPages.size(), m_listPages.size() > 0, wxPoint( 10, 30), l_size + wxSize( -20, -40));
	m_listPages.push_back( l_pReturn);
	AddPage( l_pReturn, p_strTitle);

	if (m_pSelection == NULL)
	{
		m_pSelection = l_pReturn;
	}

	if (m_pGenerator == NULL)
	{
		m_pGenerator = p_pGenerator;
	}

	return l_pReturn;
}

bool SeEffectsNotebook :: DeletePage( size_t p_uiPage)
{
	bool l_bReturn = wxNotebook::DeletePage( p_uiPage);

	if (l_bReturn)
	{
		std::vector <SeEffectPage *>::iterator l_it = m_listPages.begin();
		m_listPages.erase( l_it + p_uiPage);
		l_it = l_it + p_uiPage;

		while (l_it != m_listPages.end())
		{
			( * l_it)->DecIndex();
			++l_it;
		}
	}

	return l_bReturn;
}

BEGIN_EVENT_TABLE( SeEffectsNotebook, wxNotebook)
	EVT_SIZE(										SeEffectsNotebook::_onSize)
	EVT_NOTEBOOK_PAGE_CHANGED( SePanel::eShaders,	SeEffectsNotebook::_onSelectPage)
END_EVENT_TABLE()

void SeEffectsNotebook :: _onSize( wxSizeEvent & p_event)
{
	wxPoint l_ptPosition = wxPoint( 10, 30);
	wxSize l_size = GetClientSize() + wxSize( -20, -40);
	for (std::vector<SeEffectPage *>::iterator l_it = m_listPages.begin() ; l_it != m_listPages.end() ; ++l_it)
	{
		( * l_it)->SetSize( l_ptPosition.x, l_ptPosition.y, l_size.x, l_size.y);
	}
}

void SeEffectsNotebook :: _onSelectPage( wxBookCtrlEvent & p_event)
{
	int l_iSelection = p_event.GetSelection();

	if (l_iSelection == m_listPages.size() - 1 && m_pGenerator->EffectAdd())
	{
		wxString l_strPage = wxT( "Shader ");
		l_strPage << m_listPages.size();
		SetPageText( l_iSelection, l_strPage);
		CreatePage( m_pGenerator, wxT( "New ..."));
	}
}

//*************************************************************************************************
