#include "TextEffects/GeneratorTextEffects.h"

#include <wx/colordlg.h>
#include <wx/display.h>
#include <wx/fontdlg.h>
#include <wx/sysopt.h>
#include <wx/tglbtn.h>
#include <wx/timer.h>
#include <cmath>

#include <ft2build.h>
#include <freetype/freetype.h>
#include <freetype/ftglyph.h>
#include <freetype/ftoutln.h>
#include <freetype/fttrigon.h>

#ifdef __WXMSW__
#	include<wx/msw/msvcrt.h>
#endif

using namespace ProceduralTextures;

//*************************************************************************************************

TextEffects::Glyph :: Glyph()
{
}

TextEffects::Glyph :: Glyph( const Glyph & p_glyph)
	:	m_ptPosition	( p_glyph.m_ptPosition)
	,	m_size			( p_glyph.m_size)
	,	m_bitmap		( p_glyph.m_bitmap)
{
}

TextEffects::Glyph & TextEffects::Glyph :: operator =( const Glyph & p_glyph)
{
	m_ptPosition	= p_glyph.m_ptPosition;
	m_size			= p_glyph.m_size;
	m_bitmap		= p_glyph.m_bitmap;

	return * this;
}

TextEffects::Glyph :: ~Glyph()
{
	m_bitmap.clear();
}

//*************************************************************************************************

TextEffects :: TextEffects( int p_width, int p_height, int p_iWndId, wxFrame * p_pFrame)
	:	ProceduralGenerator( p_width, p_height, p_iWndId, p_pFrame, true)
	,	m_bFontModified		( false)
	,	m_bTextModified		( false)
	,	m_bNewTexture		( false)
	,	m_bNewTxtTexture	( false)
{
	m_blendRgbTxt.function		= GL_MODULATE;
	m_blendRgbTxt.source1		= GL_TEXTURE;
	m_blendRgbTxt.operand1		= GL_SRC_ALPHA;
	m_blendRgbTxt.source2		= GL_CONSTANT;
	m_blendRgbTxt.operand2		= GL_SRC_COLOR;
	m_blendRgbTxt.source3		= GL_NONE;
	m_blendRgbTxt.operand3		= GL_NONE;

	m_blendAlphaTxt.function	= GL_REPLACE;
	m_blendAlphaTxt.source1		= GL_TEXTURE;
	m_blendAlphaTxt.operand1	= GL_SRC_ALPHA;
	m_blendAlphaTxt.source2		= GL_NONE;
	m_blendAlphaTxt.operand2	= GL_NONE;
	m_blendAlphaTxt.source3		= GL_NONE;
	m_blendAlphaTxt.operand3	= GL_NONE;

	m_blendRgbImg.function		= GL_MODULATE;
	m_blendRgbImg.source1		= GL_PREVIOUS;
	m_blendRgbImg.operand1		= GL_ONE_MINUS_SRC_ALPHA;
	m_blendRgbImg.source2		= GL_TEXTURE;
	m_blendRgbImg.operand2		= GL_SRC_COLOR;
	m_blendRgbImg.source3		= GL_NONE;
	m_blendRgbImg.operand3		= GL_NONE;

	m_blendAlphaImg.function	= GL_REPLACE;
	m_blendAlphaImg.source1		= GL_PREVIOUS;
	m_blendAlphaImg.operand1	= GL_ONE_MINUS_SRC_ALPHA;
	m_blendAlphaImg.source2		= GL_NONE;
	m_blendAlphaImg.operand2	= GL_NONE;
	m_blendAlphaImg.source3		= GL_NONE;
	m_blendAlphaImg.operand3	= GL_NONE;

	wxSize l_size( p_width, p_height);

	m_textureImg.Resize( l_size);
	m_textureTmpImg.Resize( l_size);
	m_textureTxt.Resize( l_size);
	m_textureTmpTxt.Resize( l_size);

	m_frameBufferImg.Resize( l_size);
	m_frameBufferTxt.Resize( l_size);

	m_frameBufferImg.BindTexture( GL_COLOR_ATTACHMENT0, & m_textureTmpImg);
	m_frameBufferTxt.BindTexture( GL_COLOR_ATTACHMENT0, & m_textureTmpTxt);

	wxSystemOptions::SetOption( wxT( "msw.remap"), 0);
	wxInitAllImageHandlers();
	m_initialised = true;
}

TextEffects :: ~TextEffects()
{
	m_initialised = false;
	_cleanupArrays();
	wxImage::CleanUpHandlers();
	_cleanup();
}

void TextEffects :: SetImagePath( const wxString & p_strImagePath)
{
	m_initialised = false;
	wxSize l_size = _loadImage( p_strImagePath, m_image, m_backBuffer);
	m_frontBuffer.init( Point<size_t, 2>( l_size.x, l_size.y));
	SetSize( l_size);
	m_bNewTexture = true;
	m_initialised = true;
}

void TextEffects :: SetTxtColour( const wxColour & p_clrColour)
{
	m_txtColour.Set( p_clrColour.Red(), p_clrColour.Green(), p_clrColour.Blue(), byte( 255));
	m_textureTxt.SetColour( m_txtColour);
}

void TextEffects :: SetTxtImagePath( const wxString & p_strImagePath)
{
	if (m_txtImage.LoadFile( p_strImagePath))
	{
		_updateTxtImage();
	}
}

void TextEffects :: SetTextBackgroundType( eBG_TYPE val)
{
	m_eTextBackgroundType = val;

	if (m_eTextBackgroundType == eBG_TYPE_COLOUR)
	{
		m_blendRgbTxt.source2 = GL_CONSTANT;
	}
	else
	{
		m_blendRgbTxt.source2 = GL_PREVIOUS;
	}
}

void TextEffects :: _loadFtFont()
{
	if ( ! m_strFontFile.empty())
	{
		FT_Library l_ftLibrary;
		int l_iFontHeight = 64;
		wchar_t l_cMin = std::numeric_limits<wchar_t>::min();
		wchar_t l_cMax = std::numeric_limits<wchar_t>::max();
		m_arrayChars.resize( size_t( l_cMax) + 1, Glyph());
		m_iMaxHeight = 0;
		m_iMaxTop = 0;

		if (FT_Init_FreeType( & l_ftLibrary) == 0)
		{
			FT_Face l_ftFace;

			if (FT_New_Face( l_ftLibrary, m_strFontFile.char_str(), 0, & l_ftFace) == 0)
			{
				if (FT_Set_Char_Size( l_ftFace, l_iFontHeight << 6, l_iFontHeight << 6, 96, 96) == 0)
				{
					wxSize l_size;
					wxPoint l_ptPosition;
					std::vector<byte> l_bitmap;
					size_t l_uiIndex;
					FT_Glyph l_ftGlyph;
					int l_iTop = 0;
					size_t l_uiSize;

					for (wchar_t c = l_cMin ; c < l_cMax ; c++)
					{
						if (FT_Load_Glyph( l_ftFace, FT_Get_Char_Index( l_ftFace, c), FT_LOAD_DEFAULT) == 0)
						{
							if (FT_Get_Glyph( l_ftFace->glyph, & l_ftGlyph) == 0)
							{
								FT_Glyph_To_Bitmap( & l_ftGlyph, ft_render_mode_normal, 0, 1);
								FT_BitmapGlyph l_ftBmpGlyph = (FT_BitmapGlyph)l_ftGlyph;
								FT_Bitmap & l_ftBitmap = l_ftBmpGlyph->bitmap;

								m_iMaxTop = std::max<int>( m_iMaxTop, l_ftBmpGlyph->top);

								l_uiSize = std::abs( l_ftBitmap.pitch) * l_ftBitmap.rows;
								l_size = wxSize( std::abs( l_ftBitmap.pitch), l_ftBitmap.rows);
								l_ptPosition = wxPoint( l_ftBmpGlyph->left, l_ftBmpGlyph->top);
								l_bitmap.resize( l_uiSize, 0);
								l_uiIndex = 0;

								if (c == L' ')
								{
									l_size.x = l_iFontHeight / 2;
								}

								for (int i = 0 ; i < l_ftBitmap.rows ; i++)
								{
									for (int j = 0 ; j < std::abs( l_ftBitmap.pitch) ; j++)
									{
										l_bitmap[l_uiIndex++] = l_ftBitmap.buffer[l_ftBitmap.pitch * i + j];
									}
								}

								m_arrayChars[c].SetSize( l_size);
								m_arrayChars[c].SetPosition( l_ptPosition);
								m_arrayChars[c].SetBitmap( l_bitmap);
							}
						}
					}

					for (wchar_t c = l_cMin ; c < l_cMax ; c++)
					{
						l_ptPosition = m_arrayChars[c].GetPosition();
						l_size = m_arrayChars[c].GetSize();
						l_ptPosition = wxPoint( l_ptPosition.x, m_iMaxTop - l_ptPosition.y);
						m_arrayChars[c].SetPosition( l_ptPosition);
						m_iMaxHeight = std::max<int>( m_iMaxHeight, l_ptPosition.y + l_size.y);
					}
				}

				FT_Done_Face( l_ftFace);
			}

			FT_Done_FreeType( l_ftLibrary);
		}
	}
}

void TextEffects :: _cleanupArrays()
{
	m_arrayChars.clear();
}

void TextEffects :: _updateTxtImage()
{
	if (m_txtImage.IsOk())
	{
		wxSize l_size( m_sizeImage.x, m_sizeImage.y);
		m_txtImage.Rescale( m_sizeImage.x, m_sizeImage.y, wxIMAGE_QUALITY_HIGH);
		PixelBuffer & l_buffer = m_textureTxtImg.GetBuffer();
		l_buffer.init( Point<size_t, 2>( l_size.x, l_size.y));

		if (m_txtImage.HasAlpha())
		{
			l_buffer.set<unsigned char, 3, 1>( m_txtImage.GetData(), m_txtImage.GetAlpha());
		}
		else
		{
			l_buffer.set<unsigned char, 3, 0>( m_txtImage.GetData(), NULL);
		}

		m_bNewTxtTexture = true;
	}
}

void TextEffects :: _resize( const wxSize & p_size)
{
	ProceduralGenerator::_resize( p_size);

	m_textureImg.Resize( p_size);
	m_textureTmpImg.Resize( p_size);
	m_textureTxt.Resize( p_size);
	m_textureTmpTxt.Resize( p_size);
	m_textureTxtImg.Resize( p_size);

	m_frameBufferImg.Resize( p_size);
	m_frameBufferTxt.Resize( p_size);

	_updateTxtImage();
}

void TextEffects :: _glInit()
{
	m_textureImg.Initialise();
	m_textureTmpImg.Initialise();
	m_textureTxt.Initialise();
	m_textureTmpTxt.Initialise();
	m_textureTxtImg.Initialise();

	m_frameBufferImg.Initialise();
	m_frameBufferTxt.Initialise();
}

void TextEffects :: _glPreRender()
{
	m_textureImg.Initialise( & m_backBuffer);
	m_textureTmpImg.Initialise( m_sizeImage);
	m_textureTxt.Initialise( m_sizeImage);
	m_textureTmpTxt.Initialise( m_sizeImage);
	m_textureTxtImg.Initialise( m_sizeImage);

	m_frameBufferImg.Initialise( this);
	m_frameBufferTxt.Initialise( this);

	if (m_bFontModified)
	{
		_cleanupArrays();
		_loadFtFont();

		m_bFontModified = false;
		m_bTextModified = true;
	}

	if (m_bTextModified && m_arrayChars.size() > 0)
	{
		wxPoint l_ptPos( 0, 0);
		wxPoint l_ptCharPos;
		wxChar c;
		wxSize l_sizeChar;
		std::vector <byte> l_bitmap;
		size_t l_uiIndex = 0;
		m_textureTxt.GetBuffer().init();

		for (wxString::iterator l_it = m_strText.begin() ; l_it != m_strText.end() ; ++l_it)
		{
			c = ( * l_it);
			l_sizeChar = m_arrayChars[c].GetSize();
			l_ptCharPos = m_arrayChars[c].GetPosition();
			l_bitmap = m_arrayChars[c].GetBitmap();
			l_uiIndex = 0;

			if (c == wxT( '\n'))
			{
				l_ptPos = wxPoint( 0, l_ptPos.y + m_iMaxHeight);
			}
			else if (c == wxT( '\r'))
			{
				l_ptPos = wxPoint( 0, l_ptPos.y);
			}
			else
			{
				if (l_ptPos.x + l_ptCharPos.x + l_sizeChar.x >= m_sizeImage.x)
				{
					l_ptPos = wxPoint( 0, l_ptPos.y + m_iMaxHeight);
				}

				if (l_ptPos.y + l_ptCharPos.y + l_sizeChar.y < m_sizeImage.y)
				{
					for (int i = 0 ; i < l_sizeChar.y ; i++)
					{
						for (int j = 0 ; j < l_sizeChar.x ; j++)
						{
							byte l_r = l_bitmap[l_uiIndex++];

							if (l_r != 0)
							{
								if (l_ptPos.x + j + l_ptCharPos.x >= 0)
								{
									m_textureTxt.GetBuffer()[(m_sizeImage.y - 1) - (l_ptCharPos.y + l_ptPos.y + i)][l_ptPos.x + j + l_ptCharPos.x].Set( l_r, l_r, l_r, l_r);
								}
								else
								{
									m_textureTxt.GetBuffer()[(m_sizeImage.y - 1) - (l_ptCharPos.y + l_ptPos.y + i)][l_ptPos.x + j].Set( l_r, l_r, l_r, l_r);
								}
							}
						}
					}
				}

				l_ptPos.x += l_sizeChar.x + l_ptCharPos.x;
			}
		}

		m_textureTxt.Activate();
		m_textureTxt.UploadGlImage();
		m_textureTxt.Deactivate();
		m_bTextModified = false;
	}

	OpenGl::Viewport( 0, 0, m_sizeImage.x, m_sizeImage.y);
	OpenGl::MatrixMode( GL_PROJECTION);
	OpenGl::LoadIdentity();
	OpenGl::Ortho( 0.0, 1.0, 0.0, 1.0, 0.0, 1.0);
	OpenGl::MatrixMode( GL_MODELVIEW);
	OpenGl::LoadIdentity();
}

void TextEffects :: _glRender( bool & p_bChanged)
{
	if (m_bNewTexture)
	{
		m_textureImg.Activate();
		m_textureImg.UploadGlImage();
		m_textureImg.Deactivate();
		m_bNewTexture = false;
	}

	if (m_bNewTxtTexture)
	{
		m_textureTxtImg.Activate();
		m_textureTxtImg.UploadGlImage();
		m_textureTxtImg.Deactivate();
		m_bNewTxtTexture = false;
	}

	// First, we blend text with chosen background
	m_frameBufferTxt.Activate();
	OpenGl::Clear( GL_COLOR_BUFFER_BIT);
	if (m_eTextBackgroundType == eBG_TYPE_TEXTURE)
	{
		// Text background is a texture
		m_textureTxtImg.Activate(	GL_TEXTURE0,	GL_REPLACE,	NULL,				NULL);
		m_textureTxt.Activate(		GL_TEXTURE1,	GL_COMBINE,	& m_blendRgbTxt,	& m_blendAlphaTxt);
		_subRender();
		m_textureTxt.Deactivate(	GL_TEXTURE1);
		m_textureTxtImg.Deactivate(	GL_TEXTURE0);
	}
	else
	{
		// Text background is a colour
		m_textureTxt.Activate(		GL_TEXTURE0,	GL_COMBINE,	& m_blendRgbTxt,	& m_blendAlphaTxt);
		_subRender();
		m_textureTxt.Deactivate(	GL_TEXTURE0);
	}
	m_frameBufferTxt.Deactivate();

	// Then we cut the image where the text must be displayed
	m_frameBufferImg.Activate();
	OpenGl::Clear( GL_COLOR_BUFFER_BIT);
	OpenGl::Enable( GL_BLEND);
	m_textureTmpTxt.Activate(		GL_TEXTURE0);
	m_textureImg.Activate(			GL_TEXTURE1,	GL_COMBINE,	& m_blendRgbImg,	& m_blendAlphaImg);
	_subRender();
	m_textureImg.Deactivate(		GL_TEXTURE1);
	m_textureTmpTxt.Deactivate(		GL_TEXTURE0);
	OpenGl::Disable( GL_BLEND);
	m_frameBufferImg.Deactivate();

	// Eventually we add the two generated textures
	m_frameBuffer.Activate();
	OpenGl::Clear( GL_COLOR_BUFFER_BIT);
	OpenGl::Enable( GL_BLEND);
	m_textureTmpImg.Activate(		GL_TEXTURE0,	GL_REPLACE);
	m_textureTmpTxt.Activate(		GL_TEXTURE1,	GL_ADD);
	_subRender();
	m_textureTmpTxt.Deactivate(		GL_TEXTURE1);
	m_textureTmpImg.Deactivate(		GL_TEXTURE0);
	OpenGl::Disable( GL_BLEND);
	m_frameBuffer.Deactivate();

	p_bChanged = false;
}

void TextEffects :: _glPostRender()
{
}

void TextEffects :: _glCleanup()
{
	m_textureImg.Cleanup();
	m_textureTmpImg.Cleanup();
	m_textureTxt.Cleanup();
	m_textureTmpTxt.Cleanup();
	m_textureTxtImg.Cleanup();

	m_frameBufferImg.Cleanup();
	m_frameBufferTxt.Cleanup();
}

//******************************************************************************

IMPLEMENT_PLUGIN( TePlugin, TePanel, TextEffects)

TePlugin :: TePlugin()
{
}

TePlugin :: ~TePlugin()
{
}

wxString TePlugin :: GetName()
{
	return wxT( "Text Effects");
}

//*************************************************************************************************

TePanel :: TePanel( wxWindow * p_pWindow, wxWindowID p_id, const wxPoint & p_ptPosition, const wxSize & p_size)
	:	ConfigPanel( p_pWindow, p_id, p_ptPosition, p_size)
	,	m_pTextCaption( NULL)
{
}

TePanel :: ~TePanel()
{
}

void TePanel :: Create( TextEffects * p_pGenerator, wxWindow * p_pParent, const wxPoint & p_ptPosition, const wxSize & p_size)
{
	SetSize( p_ptPosition.x, p_ptPosition.y, p_size.x, p_size.y);
	Reparent( p_pParent);
	m_pGenerator = p_pGenerator;

	wxArrayString l_arrayChoices;
	l_arrayChoices.push_back( wxT( "Colour"));
	l_arrayChoices.push_back( wxT( "Texture"));
	m_pComboTextBackground = new wxComboBox( this, eTextBgType, wxT( "Colour"), wxPoint( 10, 0), wxSize( p_size.x - 20, 25), l_arrayChoices, wxCB_READONLY);
	m_pComboTextBackground->SetFont( GetFont());

	m_pButtonColour = new wxButton( this, eTextBg, wxT( "Text Background"), wxPoint( 10, 30), wxSize( p_size.x - 20, 25));

	m_pButtonBackground = new wxButton( this, eBackground, wxT( "Background"), wxPoint( 10, 60), wxSize( p_size.x - 20, 25));

	m_pButtonFont = new wxButton( this, eFont, wxT( "Font file"), wxPoint( 10, 90), wxSize( p_size.x - 20, 25));

	m_pTextCaption = new wxTextCtrl( this, eText, wxEmptyString, wxPoint( 10, 120), wxSize( p_size.x - 20, p_size.y - 140), wxTE_MULTILINE | wxTE_PROCESS_TAB | wxTE_WORDWRAP);
}

BEGIN_EVENT_TABLE( TePanel, ConfigPanel)
	EVT_SIZE(						TePanel::_onSize)
	EVT_BUTTON(		eBackground,	TePanel::_onBackground)
	EVT_COMBOBOX(	eTextBgType,	TePanel::_onTxtBackgroundType)
	EVT_BUTTON(		eTextBg,		TePanel::_onTxtBackground)
	EVT_BUTTON(		eFont,			TePanel::_onFont)
	EVT_TEXT(		eText,			TePanel::_onChar)
END_EVENT_TABLE()

void TePanel :: _onSize( wxSizeEvent & p_event)
{
	wxSize l_size( p_event.GetSize());

	if (m_pTextCaption != NULL)
	{
		m_pTextCaption->SetSize( wxSize( l_size.x - 20, l_size.y - 110));
	}

	p_event.Skip();
}

void TePanel :: _onTxtBackgroundType( wxCommandEvent & p_event)
{
	m_pGenerator->SetTextBackgroundType( eBG_TYPE( p_event.GetInt()));
}

void TePanel :: _onTxtBackground( wxCommandEvent & p_event)
{
	if (m_pComboTextBackground->GetSelection() == eBG_TYPE_COLOUR)
	{
		wxColourDialog l_dialog( NULL);

		if (l_dialog.ShowModal() == wxID_OK)
		{
			m_pGenerator->SetTxtColour( l_dialog.GetColourData().GetColour());
		}
	}
	else
	{
		wxString l_strWildcard = wxT( "Image files (*.bmp;*.gif;*.png;*.jpg)|*.bmp;*.gif;*.png;*.jpg|BITMAP images (*.bmp)|*.bmp|GIF images (*.gif)|*.gif|JPEG images (*.jpg)|*.jpg|PNG images (*.png)|*.png");
		wxFileDialog l_dialog( NULL, wxT( "Choose a background image"), wxEmptyString, wxEmptyString, l_strWildcard, wxFD_DEFAULT_STYLE | wxDIALOG_NO_PARENT);

		if (l_dialog.ShowModal() == wxID_OK)
		{
			m_pGenerator->SetTxtImagePath( l_dialog.GetPath());
		}
	}
}

void TePanel :: _onBackground( wxCommandEvent	& p_event)
{
	wxString l_strWildcard = wxT( "Image files (*.bmp;*.gif;*.png;*.jpg)|*.bmp;*.gif;*.png;*.jpg|BITMAP images (*.bmp)|*.bmp|GIF images (*.gif)|*.gif|JPEG images (*.jpg)|*.jpg|PNG images (*.png)|*.png");
	wxFileDialog l_dialog( NULL, wxT( "Choose a background image"), wxEmptyString, wxEmptyString, l_strWildcard, wxFD_DEFAULT_STYLE | wxDIALOG_NO_PARENT);

	if (l_dialog.ShowModal() == wxID_OK)
	{
		m_pGenerator->SetImagePath( l_dialog.GetPath());
	}
}

void TePanel :: _onFont( wxCommandEvent	& p_event)
{
	wxFileDialog l_dialog( NULL, wxT( "Choose a font file"), wxEmptyString, wxEmptyString, wxFileSelectorDefaultWildcardStr, wxFD_DEFAULT_STYLE | wxDIALOG_NO_PARENT);
	if (l_dialog.ShowModal() == wxID_OK)
	{
		m_pGenerator->SetFont( l_dialog.GetPath());
	}
}

void TePanel :: _onChar( wxCommandEvent & p_event)
{
	m_pGenerator->SetText( p_event.GetString());
}

//*************************************************************************************************